""" Splicing junction site """
from __future__ import annotations
from typing import TYPE_CHECKING, List
from moPepGen.SeqFeature import FeatureLocation
from .VariantRecord import VariantRecord


if TYPE_CHECKING:
    from moPepGen.gtf import TranscriptAnnotationModel, GenomicAnnotation
    from moPepGen.dna import DNASeqRecord

class SpliceJunction():
    """ Represents a splice junction between two genomic positions. """
    def __init__(self, upstream_start:int, upstream_end:int, downstream_start:int,
            downstream_end:int, gene_id:str, chrom:str):
        """ Constructor """
        self.upstream_start = upstream_start
        self.upstream_end = upstream_end
        self.downstream_start = downstream_start
        self.downstream_end = downstream_end
        self.gene_id = gene_id
        self.chrom = chrom

    def __str__(self) -> str:
        """ str """
        return f"<{self.__class__}: {self.upstream_start}-{self.upstream_end}:" +\
            f"{self.downstream_start}:{self.downstream_end}, "

    def align_to_transcript(self, tx_model:TranscriptAnnotationModel,
            upstream_novel:bool, downstream_novel:bool
            ) -> SpliceJunctionTranscriptAlignment:
        """ Align the junction site to a transcript.

        Args:
            - `tx_model` (TranscriptAnnotationModel): The transcript to be
              aligned to.
            - `upstream_novel` (bool): If true, the `downstream_start` must be
              matched to an exon.
            - `downstream_novel` (bool): If true, the `upstream_start` must be
              matched to an exon.
        """
        upstream_start_index = tx_model.get_exon_with_start(self.upstream_start)
        upstream_end_index = tx_model.get_exon_with_end(self.upstream_end)
        downstream_start_index = tx_model.get_exon_with_start(self.downstream_start)
        downstream_end_index = tx_model.get_exon_with_end(self.downstream_end)

        if upstream_novel and downstream_start_index == -1:
            return None

        if downstream_novel and upstream_end_index == -1:
            return None

        return SpliceJunctionTranscriptAlignment(
            junction=self,
            tx_model=tx_model,
            upstream_start_index=upstream_start_index,
            upstream_end_index=upstream_end_index,
            downstream_start_index=downstream_start_index,
            downstream_end_index=downstream_end_index,
            upstream_novel=upstream_novel,
            downstream_novel=downstream_novel
        )

    def is_novel(self, anno:GenomicAnnotation):
        """ Whether the junction is novel, i.e. not present in at least one
        transcript isoform of the gene. """
        gene_model = anno.genes[self.gene_id]
        junction = FeatureLocation(
            start=self.upstream_end, end=self.downstream_start
        )
        for tx_id in gene_model.transcripts:
            tx_model = anno.transcripts[tx_id]
            if tx_model.has_junction(junction):
                return False
        return True

class SpliceJunctionTranscriptAlignment():
    """ Alignment between a splice junction to a transcript. """
    def __init__(self, junction:SpliceJunction, tx_model:TranscriptAnnotationModel,
            upstream_start_index:int, upstream_end_index:int,
            downstream_start_index:int, downstream_end_index:int,
            upstream_novel:bool, downstream_novel:bool):
        """ Constructor """
        self.junction = junction
        self.tx_model = tx_model
        self.upstream_start_index = upstream_start_index
        self.upstream_end_index = upstream_end_index
        self.downstream_start_index = downstream_start_index
        self.downstream_end_index = downstream_end_index
        self.upstream_novel = upstream_novel
        self.downstream_novel = downstream_novel

    def get_interjacent_exons(self) -> List[int]:
        """ Get interjacent exons between upstream end and downstream start. """
        if not self.upstream_end_index and not self.downstream_start_index:
            raise ValueError(
                "Neither upstream end nor downstream start has a matched exon."
            )

        interjacent = []
        if self.upstream_end_index > -1:
            if self.upstream_end_index + 1 == len(self.tx_model.exon):
                return interjacent
            iter_exon = range(self.upstream_end_index + 1, len(self.tx_model.exon))
            is_reversed = False
        else:
            if self.downstream_start_index == 0:
                return interjacent
            iter_exon = range(self.downstream_start_index - 1, -1, -1)
            is_reversed = True

        for i in iter_exon:
            exon = self.tx_model.exon[i]
            if self.junction.upstream_end <= exon.location.start \
                    < exon.location.end <= self.junction.downstream_start:
                interjacent.append(i)
            if exon.location.end <= self.junction.upstream_end \
                    or exon.location.start >= self.junction.downstream_start:
                break

        if is_reversed:
            interjacent = list(reversed(interjacent))
        return interjacent

    def get_upstream_end_spanning(self) -> int:
        """ Get exons that are spanning over the upstream end position. """
        if self.downstream_start_index == -1:
            return self.tx_model.get_exon_containing(self.junction.upstream_end - 1)

        i = self.downstream_start_index - 1
        while i >= 0:
            exon = self.tx_model.exon[i]
            if self.junction.upstream_end - 1 in exon.location:
                return i
            i -= 1
        return -1

    def get_downstream_start_spanning(self) -> int:
        """ Get exons that are spanning over the downstream start position. """
        if self.upstream_end_index == -1:
            return self.tx_model.get_exon_containing(self.junction.downstream_start)

        i = self.upstream_end_index + 1
        while i < len(self.tx_model.exon):
            exon = self.tx_model.exon[i]
            if self.junction.downstream_start in exon.location:
                return i
            i += 1
        return -1

    def create_upstream_deletion(self, spanning:int, interjacent:List[int],
            anno:GenomicAnnotation, gene_seq:DNASeqRecord, var_id:str
            ) -> VariantRecord:
        """ Create an upstream deletion variant. """
        gene_id = self.junction.gene_id
        gene_model = anno.genes[gene_id]
        tx_id = self.tx_model.transcript.transcript_id
        chrom = gene_model.chrom
        strand = gene_model.strand
        tx_model = self.tx_model

        if tx_model.exon[spanning].location.end == self.junction.upstream_end:
            genomic_start = tx_model.exon[interjacent[0]].location.start
        else:
            genomic_start = self.junction.upstream_end
        start = anno.coordinate_genomic_to_gene(genomic_start, gene_id)

        if interjacent:
            genomic_end = tx_model.exon[interjacent[-1]].location.end
        else:
            genomic_end = tx_model.exon[spanning].location.end
        end = anno.coordinate_genomic_to_gene(genomic_end - 1, gene_id)

        if strand == -1:
            start, end = end, start
        end += 1

        ref = str(gene_seq.seq[start])

        genomic_position = f'{chrom}:{genomic_start + 1}:{genomic_end}'

        location = FeatureLocation(seqname=gene_id, start=start, end=end)
        alt = '<DEL>'
        attrs = {
            'TRANSCRIPT_ID': tx_id,
            'START': start,
            'END': end,
            'GENE_SYMBOL': gene_model.gene_name,
            'GENOMIC_POSITION': genomic_position
        }
        _type = 'Deletion'
        return VariantRecord(location, ref, alt, _type, var_id, attrs)

    def create_downstream_deletion(self, spanning:int, interjacent:List[int],
            anno:GenomicAnnotation, gene_seq:DNASeqRecord, var_id:str
            ) -> VariantRecord:
        """ Create a downstream deletion variant. """
        gene_id = self.junction.gene_id
        gene_model = anno.genes[gene_id]
        tx_id = self.tx_model.transcript.transcript_id
        chrom = gene_model.chrom
        strand = gene_model.strand
        tx_model = self.tx_model

        if interjacent:
            genomic_start = tx_model.exon[interjacent[0]].location.start
        else:
            genomic_start = tx_model.exon[spanning].location.start
        start = anno.coordinate_genomic_to_gene(genomic_start, gene_id)

        if self.junction.downstream_start == tx_model.exon[spanning].location.start:
            genomic_end = tx_model.exon[interjacent[-1]].location.end
        else:
            genomic_end = self.junction.downstream_start
        end = anno.coordinate_genomic_to_gene(genomic_end - 1, gene_id)

        if strand == -1:
            start, end = end, start
        end += 1

        ref = str(gene_seq.seq[start])

        genomic_position = f'{chrom}:{genomic_start + 1}:{genomic_end}'

        location = FeatureLocation(seqname=gene_id, start=start, end=end)
        alt = '<DEL>'
        attrs = {
            'TRANSCRIPT_ID': tx_id,
            'START': start,
            'END': end,
            'GENE_SYMBOL': gene_model.gene_name,
            'GENOMIC_POSITION': genomic_position
        }
        _type = 'Deletion'
        return VariantRecord(location, ref, alt, _type, var_id, attrs)

    def create_upstream_substitution(self, interjacent:List[int],
            anno:GenomicAnnotation, gene_seq:DNASeqRecord, var_id:str
            ) -> VariantRecord:
        """ Create an upstream substitution variant. """
        gene_id = self.junction.gene_id
        gene_model = anno.genes[gene_id]
        tx_id = self.tx_model.transcript.transcript_id
        chrom = gene_model.chrom
        strand = gene_model.strand
        tx_model = self.tx_model

        genomic_start = tx_model.exon[interjacent[0]].location.start
        start = anno.coordinate_genomic_to_gene(genomic_start, gene_id)
        genomic_end = tx_model.exon[interjacent[-1]].location.end
        end = anno.coordinate_genomic_to_gene(genomic_end - 1, gene_id)

        if interjacent[0] > 0:
            genomic_donor_start = tx_model.exon[interjacent[0] - 1].location.end
            if self.junction.upstream_start is not None:
                genomic_donor_start = max(genomic_donor_start, self.junction.upstream_start)
        else:
            if self.junction.upstream_start is None:
                raise ValueError('`upstream_start` is missing.')
            genomic_donor_start = self.junction.upstream_start
        donor_start = anno.coordinate_genomic_to_gene(genomic_donor_start, gene_id)
        genomic_donor_end = self.junction.upstream_end
        donor_end = anno.coordinate_genomic_to_gene(genomic_donor_end - 1, gene_id)

        if strand == -1:
            start, end = end, start
            donor_start, donor_end = donor_end, donor_start

        end += 1
        donor_end += 1

        ref = str(gene_seq.seq[start])
        genomic_position = f'{chrom}:{genomic_start + 1}:{genomic_end}'

        location = FeatureLocation(seqname=gene_id, start=start, end=end)
        alt = '<SUB>'
        attrs = {
            'TRANSCRIPT_ID': tx_id,
            'START': start,
            'END': end,
            'DONOR_START': donor_start,
            'DONOR_END': donor_end,
            'DONOR_GENE_ID': gene_id,
            'GENE_SYMBOL': gene_model.gene_name,
            'GENOMIC_POSITION': genomic_position
        }
        _type = 'Substitution'
        return VariantRecord(location, ref, alt, _type, var_id, attrs)

    def create_downstream_substitution(self, interjacent:List[int],
            anno:GenomicAnnotation, gene_seq:DNASeqRecord, var_id:str
            ) -> VariantRecord:
        """ Create a downstream substitution variant. """
        gene_id = self.junction.gene_id
        gene_model = anno.genes[gene_id]
        tx_id = self.tx_model.transcript.transcript_id
        chrom = gene_model.chrom
        strand = gene_model.strand
        tx_model = self.tx_model

        genomic_start = tx_model.exon[interjacent[0]].location.start
        start = anno.coordinate_genomic_to_gene(genomic_start, gene_id)
        genomic_end = tx_model.exon[interjacent[-1]].location.end
        end = anno.coordinate_genomic_to_gene(genomic_end - 1, gene_id)

        genomic_donor_start = self.junction.downstream_start
        donor_start = anno.coordinate_genomic_to_gene(genomic_donor_start, gene_id)

        if interjacent[-1] < len(tx_model.exon) - 1:
            genomic_donor_end = tx_model.exon[interjacent[-1] + 1].location.start
            if self.junction.downstream_end is not None:
                genomic_donor_end = min(genomic_donor_end, self.junction.downstream_end)
        else:
            if self.junction.downstream_end is None:
                raise ValueError('`downstream_start` is missing.')
            genomic_donor_end = self.junction.downstream_end
        donor_end = anno.coordinate_genomic_to_gene(genomic_donor_end - 1, gene_id)

        if strand == -1:
            start, end = end, start
            donor_start, donor_end = donor_end, donor_start

        end += 1
        donor_end += 1

        ref = str(gene_seq.seq[start])
        genomic_position = f'{chrom}:{genomic_start + 1}:{genomic_end}'

        location = FeatureLocation(seqname=gene_id, start=start, end=end)
        alt = '<SUB>'
        attrs = {
            'TRANSCRIPT_ID': tx_id,
            'START': start,
            'END': end,
            'DONOR_START': donor_start,
            'DONOR_END': donor_end,
            'DONOR_GENE_ID': gene_id,
            'GENE_SYMBOL': gene_model.gene_name,
            'GENOMIC_POSITION': genomic_position
        }
        _type = 'Substitution'
        return VariantRecord(location, ref, alt, _type, var_id, attrs)

    def create_upstream_insertion(self, anno:GenomicAnnotation,
            gene_seq:DNASeqRecord, var_id:str) -> VariantRecord:
        """ Create an upstream insertion variant. """
        gene_id = self.junction.gene_id
        gene_model = anno.genes[gene_id]
        tx_id = self.tx_model.transcript.transcript_id
        chrom = gene_model.chrom
        strand = gene_model.strand
        tx_model = self.tx_model

        if self.downstream_start_index <= 0:
            raise ValueError("`downstream_start_index` should not be > 0.")

        if strand == 1:
            insert_pos_genomic = tx_model.exon[self.downstream_start_index - 1].location.end - 1
            donor_start = tx_model.exon[self.downstream_start_index - 1].location.end
            if self.junction.downstream_end is not None:
                donor_start = max(donor_start, self.junction.upstream_start)
            donor_end = self.junction.upstream_end - 1
        else:
            insert_pos_genomic = self.junction.downstream_start
            donor_start = self.junction.upstream_end - 1
            donor_end = tx_model.exon[self.downstream_start_index - 1].location.end
            if self.junction.upstream_start is not None:
                donor_end = max(donor_end, self.junction.upstream_start)

        insert_pos = anno.coordinate_genomic_to_gene(insert_pos_genomic, gene_id)

        donor_start = anno.coordinate_genomic_to_gene(donor_start, gene_id)
        donor_end = anno.coordinate_genomic_to_gene(donor_end, gene_id) + 1

        ref = str(gene_seq.seq[insert_pos])
        genomic_position = f'{chrom}:{insert_pos_genomic + 1}:{insert_pos_genomic + 2}'

        location = FeatureLocation(
            seqname=gene_id,
            start=insert_pos,
            end=insert_pos + 1
        )
        alt = '<Ins>'
        attrs = {
            'TRANSCRIPT_ID': tx_id,
            'DONOR_GENE_ID': gene_id,
            'DONOR_START': donor_start,
            'DONOR_END': donor_end,
            'GENE_SYMBOL': gene_model.gene_name,
            'GENOMIC_POSITION': genomic_position
        }
        _type = 'Insertion'
        return VariantRecord(location, ref, alt, _type, var_id, attrs)

    def create_downstream_insertion(self, anno:GenomicAnnotation,
            gene_seq:DNASeqRecord, var_id:str) -> VariantRecord:
        """ Create a downstream insertion variant. """
        gene_id = self.junction.gene_id
        gene_model = anno.genes[gene_id]
        tx_id = self.tx_model.transcript.transcript_id
        chrom = gene_model.chrom
        strand = gene_model.strand
        tx_model = self.tx_model

        if self.upstream_end_index == -1:
            raise ValueError("`upstream_start_index` should not be -1.")

        if self.upstream_start_index == len(tx_model.exon) - 1:
            raise ValueError("`downstream_start_index` should not be the last one.")

        if strand == 1:
            insert_pos_genomic = self.junction.upstream_end - 1
            donor_start = self.junction.downstream_start
            donor_end = tx_model.exon[self.upstream_end_index + 1].location.start - 1
            if self.junction.downstream_end is not None:
                donor_end = min(donor_end, self.junction.downstream_end - 1)
        else:
            insert_pos_genomic = tx_model.exon[self.upstream_end_index + 1].location.start
            donor_start = tx_model.exon[self.upstream_end_index + 1].location.start - 1
            if self.junction.downstream_end is not None:
                donor_start = min(donor_start, self.junction.downstream_end - 1)
            donor_end = self.junction.downstream_start
        insert_pos = anno.coordinate_genomic_to_gene(insert_pos_genomic, gene_id)

        donor_start = anno.coordinate_genomic_to_gene(donor_start, gene_id)
        donor_end = anno.coordinate_genomic_to_gene(donor_end, gene_id) + 1

        ref = str(gene_seq.seq[insert_pos])
        genomic_position = f'{chrom}:{insert_pos_genomic + 1}:{insert_pos_genomic + 2}'

        location = FeatureLocation(
            seqname=gene_id,
            start=insert_pos,
            end=insert_pos + 1
        )
        alt = '<Ins>'
        attrs = {
            'TRANSCRIPT_ID': tx_id,
            'DONOR_GENE_ID': gene_id,
            'DONOR_START': donor_start,
            'DONOR_END': donor_end,
            'GENE_SYMBOL': gene_model.gene_name,
            'GENOMIC_POSITION': genomic_position
        }
        _type = 'Insertion'
        return VariantRecord(location, ref, alt, _type, var_id, attrs)

    def convert_to_variant_records(self, anno:GenomicAnnotation,
            gene_seq:DNASeqRecord, var_id:str):
        """ Convert splicing to variant records. """
        variants = []
        interjacent = self.get_interjacent_exons()

        if self.upstream_novel:
            if self.upstream_end_index == -1 or len(interjacent) > 0:
                spanning = self.get_upstream_end_spanning()
                if spanning > -1:
                    v = self.create_upstream_deletion(
                        spanning, interjacent, anno, gene_seq, var_id
                    )
                    variants.append(v)
                elif len(interjacent) > 0:
                    v = self.create_upstream_substitution(
                        interjacent, anno, gene_seq, var_id
                    )
                    variants.append(v)
                elif self.downstream_start_index > 0:
                    v = self.create_upstream_insertion(
                        anno, gene_seq, var_id
                    )
                    variants.append(v)

        if self.downstream_novel:
            if self.downstream_start_index == -1 or len(interjacent) > 0:
                spanning = self.get_downstream_start_spanning()
                if spanning > -1:
                    v = self.create_downstream_deletion(
                        spanning, interjacent, anno, gene_seq, var_id
                    )
                    variants.append(v)
                elif len(interjacent) > 0:
                    v = self.create_downstream_substitution(
                        interjacent, anno, gene_seq, var_id
                    )
                    variants.append(v)
                elif -1 < self.downstream_end_index < len(self.tx_model.exon) - 1:
                    v = self.create_downstream_insertion(
                        anno, gene_seq, var_id
                    )
                    variants.append(v)

        if not self.downstream_novel and not self.upstream_novel:
            strand = self.tx_model.transcript.strand
            is_upstream_aligned = (strand == 1 and self.upstream_end_index != -1) \
                    or (strand == -1 and self.downstream_start_index != -1)
            tx_start = self.tx_model.transcript.location.start
            tx_end = self.tx_model.transcript.location.end
            is_after_tx_start = \
                (strand == 1 and tx_start < self.junction.upstream_end) \
                or (strand == -1 and tx_end > self.junction.downstream_start + 1)
            if is_upstream_aligned and is_after_tx_start:
                if strand == 1:
                    if self.downstream_start_index == -1 or len(interjacent) > 0:
                        spanning = self.get_downstream_start_spanning()
                        if spanning > -1:
                            v = self.create_downstream_deletion(
                                spanning, interjacent, anno, gene_seq, var_id
                            )
                            variants.append(v)
                        elif len(interjacent) > 0:
                            v = self.create_downstream_substitution(
                                interjacent, anno, gene_seq, var_id
                            )
                            variants.append(v)
                else:
                    spanning = self.get_upstream_end_spanning()
                    if self.upstream_end_index == -1 or len(interjacent) > 0:
                        if spanning > -1:
                            v = self.create_upstream_deletion(
                                spanning, interjacent, anno, gene_seq, var_id
                            )
                            variants.append(v)
                        elif len(interjacent) > 0:
                            v = self.create_upstream_substitution(
                                interjacent, anno, gene_seq, var_id
                            )
                            variants.append(v)

        return variants



""" `parseArriba` takes the identified fusion transcript results from
[Arriba](https://github.com/suhrig/arriba) and saves as a GVF file. The GVF
file can be later used to call variant peptides using
[callVariant](call-variant.md)."""
from typing import List
from pathlib import Path
import argparse
from moPepGen import get_logger, seqvar, parser, err
from moPepGen.cli import common


INPUT_FILE_FORMATS = ['.tsv', '.txt']
OUTPUT_FILE_FORMATS = ['.gvf']

# pylint: disable=W0212
def add_subparser_parse_arriba(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen parseArriba """

    p:argparse.ArgumentParser = subparsers.add_parser(
        name='parseArriba',
        help='Parse Arriba result for moPepGen to call variant peptides.',
        description='Parse the Arriba result to GVF format of variant'
        ' records for moPepGen to call variant peptides.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=p, formats=INPUT_FILE_FORMATS,
        message="File path to Arriba's output TSV file."
    )
    common.add_args_output_path(p, OUTPUT_FILE_FORMATS)
    p.add_argument(
        '--min-split-read1',
        type=int,
        help='Minimal split_read1 value.',
        metavar='<value>',
        default=1
    )
    p.add_argument(
        '--min-split-read2',
        type=int,
        help='Minimal split_read2 value.',
        metavar='<value>',
        default=1
    )
    p.add_argument(
        '--min-confidence',
        type=str,
        choices=parser.ArribaParser.ArribaConfidence.levels.keys(),
        help='Minimal confidence value.',
        metavar='<choice>',
        default='medium'
    )
    common.add_args_source(p)
    common.add_args_reference(p, proteome=False)
    common.add_args_debug_level(p)
    p.set_defaults(func=parse_arriba)
    common.print_help_if_missing_args(p)
    return p

def parse_arriba(args:argparse.Namespace) -> None:
    """ Parse Arriba output and save it in GVF format. """
    logger = get_logger()
    # unpack args
    fusion = args.input_path
    output_path:Path = args.output_path
    common.validate_file_format(
        fusion, INPUT_FILE_FORMATS, check_readable=True
    )
    common.validate_file_format(
        output_path, OUTPUT_FILE_FORMATS, check_writable=True
    )

    min_split_read1:int = args.min_split_read1
    min_split_read2:int = args.min_split_read2
    min_confidence:str = args.min_confidence

    common.print_start_message(args)

    genome, anno, *_ = common.load_references(args=args, load_canonical_peptides=False)

    variants:List[seqvar.VariantRecord] = []

    with open(fusion, 'rt') as handle:
        for record in parser.ArribaParser.parse(handle):
            if not record.gene_id1 in anno.genes or not record.gene_id2 in anno.genes:
                continue
            if not record.is_valid(min_split_read1, min_split_read2, min_confidence):
                continue
            if record.transcript_on_antisense_strand(anno):
                continue
            try:
                var_records = record.convert_to_variant_records(anno, genome)
            except err.GeneNotFoundError:
                continue
            variants.extend(var_records)

    logger.info('Arriba output %s loaded.', fusion)

    if not variants:
        logger.warning('No variant record is saved.')
        return

    genes_rank = anno.get_genes_rank()
    variants = sorted(variants, key=lambda x: genes_rank[x.location.seqname])

    logger.info('Variants sorted.')

    metadata = common.generate_metadata(args)

    seqvar.io.write(variants, output_path, metadata)

    logger.info("Variants written to disk.")



""" Module for GTFSeqFeature """
from __future__ import annotations
import re
from moPepGen.SeqFeature import SeqFeature


class GTFSeqFeature(SeqFeature):
    """ GTF specific SeqFeture """
    def __init__(self, *args, source:str=None, frame:int=None, **kwargs):
        """ Constructor """
        super().__init__(*args, **kwargs)
        self.source = source
        self.frame = frame

    @property
    def biotype(self) -> str:
        """ biotype """
        if self.source == 'GENCODE':
            return self.attributes['gene_type']
        if self.source == 'ENSEMBL':
            return self.attributes['gene_biotype']
        raise ValueError(f'Annotation source {self.source} not supported')

    @biotype.setter
    def biotype(self, biotype:str) -> None:
        """ biotype """
        if self.source == 'GENCODE':
            self.attributes['gene_type'] = biotype
        elif self.source == 'ENSEMBL':
            self.attributes['gene_biotype'] = biotype
        else:
            raise ValueError(f'Annotation source {self.source} not supported')

    @property
    def gene_id(self) -> str:
        """ gene ID """
        if 'gene_id' not in self.attributes:
            return None
        return self.attributes['gene_id']

    @gene_id.setter
    def gene_id(self, gene_id:str) -> None:
        """ set gene ID """
        self.attributes['gene_id'] = gene_id

    @property
    def transcript_id(self) -> str:
        """ transcript ID """
        if 'transcript_id' not in self.attributes:
            return None
        return self.attributes['transcript_id']

    @transcript_id.setter
    def transcript_id(self, transcript_id:str) -> None:
        """ set transcript ID """
        self.attributes['transcript_id'] = transcript_id

    @property
    def protein_id(self) -> str:
        """ protein ID """
        if 'protein_id' not in self.attributes:
            return None
        return self.attributes['protein_id']

    @protein_id.setter
    def protein_id(self, protein_id:str) -> None:
        """ set protein ID """
        self.attributes['protein_id'] = protein_id

    @property
    def gene_name(self) -> str:
        """ gene name """
        if 'gene_name' in self.attributes:
            return self.attributes['gene_name']
        return ''

    @gene_name.setter
    def gene_name(self, gene_name:str) -> None:
        """ set gene name """
        self.attributes['gene_name'] = gene_name

    def _shift(self, offset:int) -> GTFSeqFeature:
        """ shift by i """
        new_feature = super()._shift(offset)
        new_feature.__class__ = self.__class__
        return new_feature

    def infer_annotation_source(self):
        """ Infer the annotation source from a GTF record. Returns GENCODE,
        ENSEMBL, or None. """
        pattern = re.compile('chr[0-9XYM]{1,2}')
        if pattern.search(self.chrom):
            self.source = 'GENCODE'
            return

        pattern = re.compile(r'[A-Z]{2}[0-9]{6}\.[0-9]{1}')
        if pattern.search(self.chrom):
            self.source = 'GENCODE'
            return

        pattern = re.compile('[0-9]{1,2}')
        if pattern.search(self.chrom):
            self.source = 'ENSEMBL'
            return

        if self.chrom in ['X', 'Y', 'MT']:
            self.source = 'ENSEMBL'
            return

        pattern = re.compile('^CHR_H.+')
        if pattern.search(self.chrom):
            self.source = 'ENSEMBL'



""" ResourcesMonitor monitors the resources usage including CPU and memory """
from __future__ import annotations
import datetime
import math
import psutil


class MemoryUnit():
    """ Memory Unit """
    _suffixes = ['bytes', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB']
    def __init__(self, size:int):
        """ constructor """
        self._data = size

    def __str__(self) -> str:
        """ str """
        return self.human_readable(self._data)

    def human_readable(self, x:int) -> str:
        """ Get humand readable representation of the memroy unit. """
        # https://stackoverflow.com/a/25613067/11081630
        order = int(math.log2(x) / 10) if x else 0
        val = x/(1 << (order * 10))
        unit = self._suffixes[order]
        return f"{val:.4g} {unit}"

class ResourcesMonitor():
    """ Monitor the resources usage of the current process. """
    def __init__(self):
        """ constructor """
        self.process = psutil.Process()
        # According to the documentation, this function returns the cpu percent
        # since the last call. So calling it at the beginning.
        self.process.cpu_percent()

    def get_resource_usage(self) -> ResourcesUsage:
        """ Get resources usage """
        mem = self.process.memory_info()
        rss = MemoryUnit(mem.rss)
        vms = MemoryUnit(mem.vms)

        cpu_times = self.process.cpu_times()
        wallclock = datetime.timedelta(seconds=cpu_times.user)
        system = datetime.timedelta(seconds=cpu_times.system)

        cpu_percent = self.process.cpu_percent()

        return ResourcesUsage(
            rss=rss, vms=vms, wallclock=wallclock, system=system,
            cpu_percent=cpu_percent
        )

class ResourcesUsage():
    """ Resources usage """
    def __init__(self, rss:MemoryUnit=None, vms:MemoryUnit=None,
            wallclock:datetime.timedelta=None, system:datetime.timedelta=None,
            cpu_percent:float=None):
        """ """
        self.rss = rss
        self.vms = vms
        self.wallclock = wallclock
        self.system = system
        self.cpu_percent = cpu_percent

    def __str__(self) -> str:
        """ str """
        return  f"rss={self.rss}, vms={self.vms}, wallclock={self.wallclock}, " +\
            f"system={self.system}, cpu_usage={self.cpu_percent}%"



""" `MergeFasta` merges a serious of database FASTA files into one. This is
useful when working with multiplexed proteomic experiments such as TMT """
from __future__ import annotations
import argparse
from moPepGen import get_logger
from moPepGen.aa.VariantPeptidePool import VariantPeptidePool
from moPepGen.cli import common


INPUT_FILE_FORMATS = ['.fasta', '.fa']
OUTPUT_FILE_FORMATS = ['.fasta', '.fa']

# pylint: disable=W0212
def add_subparser_merge_fasta(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen MergeFasta """
    parser:argparse.ArgumentParser = subparsers.add_parser(
        name='mergeFasta',
        help='Merge multiple variant peptide FASTA files into one.',
        description='Merge multiple variant peptide FASTA files into one.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=parser, formats=INPUT_FILE_FORMATS, plural=True,
        message='File path to the variant peptide FASTA files.'
    )
    common.add_args_output_path(
        parser=parser, formats=OUTPUT_FILE_FORMATS
    )
    parser.add_argument(
        '--dedup-header',
        action='store_true',
        help='Remove duplicate FASTA header entries after merging.'
    )
    common.add_args_debug_level(parser)
    parser.set_defaults(func=merge_fasta)
    common.print_help_if_missing_args(parser)
    return parser

def merge_fasta(args:argparse.Namespace):
    """ Merge mulitple variant peptide FASTA files into one. """
    logger = get_logger()
    input_files = args.input_path
    for file in input_files:
        common.validate_file_format(
            file, INPUT_FILE_FORMATS, check_readable=True
        )
    output_file = args.output_path
    common.validate_file_format(
        output_file, OUTPUT_FILE_FORMATS, check_writable=True
    )

    pool = None

    for file in input_files:
        with open(file, 'rt') as handle:
            if pool is None:
                pool = VariantPeptidePool.load(handle)
            else:
                second_pool = VariantPeptidePool.load(handle)
                for peptide in second_pool.peptides:
                    pool.add_peptide(
                        peptide=peptide,
                        canonical_peptides=set(),
                        skip_checking=True
                    )
            logger.info("Database FASTA file loaded: %s", file)

    if args.dedup_header:
        pool.remove_redundant_headers()
    pool.write(output_file)

    logger.info("Merged FASTA file saved to %s", output_file)



""" `callVariant` is the core of moPepGen. It takes multiple GVF files, generated
by any moPepGen parser, and calls variant peptides caused by genomic variants
using a graph-based algorithm. For any transcript, it creates a three-frame
transcript variant graph by incorporating all variants from any sources (SNV,
INDEL, fusion, alternative splicing, RNA editing, and circRNA). The transcript
variant graph is then translated into a peptide variant graph, followed by
converting to a cleavage graph based on the enzymatic cleavage rule. The
variant peptide graph is than used to call for variant peptides that contains
at least one variant, and do not present in the canonical peptide pool.
 """
from __future__ import annotations
import argparse
import copy
import json
from typing import List, Set, TYPE_CHECKING, Dict, Tuple, Union
from pathlib import Path
from contextlib import ExitStack
from pathos.pools import ParallelPool
from moPepGen import svgraph, aa, seqvar, gtf, params, get_logger
from moPepGen.cli import common
from moPepGen.SeqFeature import FeatureLocation, SeqFeature


if TYPE_CHECKING:
    from logging import Logger
    from Bio.Seq import Seq
    from moPepGen import dna, circ
    from moPepGen.svgraph import ThreeFrameCVG, ThreeFrameTVG, PeptideVariantGraph
    from moPepGen.svgraph.VariantPeptideDict import AnnotatedPeptideLabel


INPUT_FILE_FORMATS = ['.gvf']
OUTPUT_FILE_FORMATS = ['.fasta', '.fa']

# pylint: disable=W0212
def add_subparser_call_variant(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen callPeptides """
    p:argparse.ArgumentParser = subparsers.add_parser(
        name='callVariant',
        help='Call non-canonical peptides from genomic variants.',
        description='Genomic variant data must be generated by one of the'
        ' moPepGen parser. See moPepGen --help',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    common.add_args_input_path(
        parser=p,
        formats=INPUT_FILE_FORMATS,
        plural=True,
        required=True,
        message='File path to GVF files. Must be generated by any of the'
        ' moPepGen parsers.'
    )
    common.add_args_output_path(p, OUTPUT_FILE_FORMATS)
    p.add_argument(
        '--graph-output-dir',
        type=Path,
        required=False,
        default=None,
        help='Directory path that graph data are saved to. Graph data are not'
        ' saved if this is not given.',
        metavar='<file>'
    )
    p.add_argument(
        '--max-adjacent-as-mnv',
        type=int,
        help='Max number of adjacent variants that should be merged.',
        default=2
    )
    p.add_argument(
        '--selenocysteine-termination',
        action='store_true',
        help='Include peptides of selenoprotiens that the UGA is treated as '
        'termination instead of Sec.'
    )
    p.add_argument(
        '--w2f-reassignment',
        action='store_true',
        help='Include peptides with W > F (Tryptophan to Phenylalanine) '
        'reassignment.'
    )
    p.add_argument(
        '--backsplicing-only',
        action='store_true',
        help='For circRNA, only keep noncanonical peptides spaning the backsplicing site.'
    )
    p.add_argument(
        '--coding-novel-orf',
        action='store_true',
        help='Find alternative start site for coding transcripts.'
    )
    p.add_argument(
        '--max-variants-per-node',
        type=int,
        help='Maximal number of variants per node. This argument can be useful'
        ' when there are local regions that are heavily mutated. When creating'
        ' the cleavage graph, nodes containing variants larger than this value'
        ' are skipped. Setting to -1 will avoid checking for this.',
        default=(7,),
        nargs='+',
        metavar='<number>'
    )
    p.add_argument(
        '--additional-variants-per-misc',
        type=int,
        help='Additional variants allowed for every miscleavage. This argument'
        ' is used together with --max-variants-per-node to handle hypermutated'
        ' regions. Setting to -1 will avoid checking for this.',
        default=(2,),
        nargs='+',
        metavar='<number>'
    )
    p.add_argument(
        '--min-nodes-to-collapse',
        type=int,
        help='When making the cleavage graph, the minimal number of nodes'
        ' to trigger pop collapse.',
        default=30,
        metavar='<number>'
    )
    p.add_argument(
        '--naa-to-collapse',
        type=int,
        help='The number of bases used for pop collapse.',
        default=5,
        metavar='<number>'
    )
    p.add_argument(
        '--noncanonical-transcripts',
        action='store_true',
        help='Process only noncanonical transcripts of fusion transcripts and'
        ' circRNA. Canonical transcripts are skipped.'
    )
    p.add_argument(
        '--timeout-seconds',
        type=int,
        default=1800,
        help='Time out in seconds for each transcript.'
    )
    p.add_argument(
        '--threads',
        type=int,
        help='Set number of threads to be used.',
        default=1,
        metavar='<number>'
    )
    common.add_args_reference(p)
    common.add_args_cleavage(p)
    common.add_args_debug_level(p)

    p.set_defaults(func=call_variant_peptide)
    common.print_help_if_missing_args(p)
    return p


class TallyTable:
    """ Tally table for callVariant """
    def __init__(self, logger:Logger):
        """ constructor """
        self.n_transcripts_total = 0
        self.n_transcripts_processed = 0
        self.n_total_peptides = 0
        self.n_valid_peptides = 0
        self.logger = logger

    def log(self):
        """ log """
        self.logger.info("callVariant summary:")
        self.logger.info("Total transcripts with variants: %i", self.n_transcripts_total)
        self.logger.info("Total transcripts processed: %i", self.n_transcripts_processed)
        self.logger.info(
            "Total variant peptides generated (including redundant): %i",
            self.n_total_peptides
        )
        self.logger.info("Total variant peptides saved: %i", self.n_valid_peptides)

class VariantPeptideCaller():
    """ Helper class to call variant peptides """
    def __init__(self, args:argparse.Namespace):
        """ constructor """
        self.args = args
        self.variant_files:List[Path] = args.input_path

        for file in self.variant_files:
            common.validate_file_format(
                file, INPUT_FILE_FORMATS, check_readable=True
            )

        self.output_path:str = args.output_path
        common.validate_file_format(
            self.output_path, OUTPUT_FILE_FORMATS, check_writable=True
        )
        self.peptide_table_output_path:Path = \
            Path(args.output_path).parent/f"{Path(args.output_path).stem}_peptide_table.txt"
        self.graph_output_dir:Path = args.graph_output_dir
        if self.graph_output_dir is not None:
            common.validate_file_format(
                self.graph_output_dir, check_writable=True, is_directory=True
            )

        self.cleavage_params = params.CleavageParams(
            enzyme=args.cleavage_rule,
            exception=args.cleavage_exception,
            miscleavage=int(args.miscleavage),
            min_mw=float(args.min_mw),
            min_length=args.min_length,
            max_length=args.max_length,
            max_variants_per_node = args.max_variants_per_node[0],
            additional_variants_per_misc = args.additional_variants_per_misc[0],
            min_nodes_to_collapse = args.min_nodes_to_collapse,
            naa_to_collapse = args.naa_to_collapse
        )
        self.max_adjacent_as_mnv = args.max_adjacent_as_mnv
        self.truncate_sec:bool = args.selenocysteine_termination
        self.backsplicing_only:bool = args.backsplicing_only
        self.w2f_reassignment = args.w2f_reassignment
        self.noncanonical_transcripts = args.noncanonical_transcripts
        self.invalid_protein_as_noncoding:bool = args.invalid_protein_as_noncoding
        self.threads = args.threads
        self.reference_data = None
        self.variant_peptides = aa.VariantPeptidePool()
        self.variant_record_pool:seqvar.VariantRecordPoolOnDisk = None

        if not self.variant_files:
            if not (self.w2f_reassignment or self.truncate_sec):
                raise ValueError('Please provide at least a GVF file.')
            self.process_all_transcripts = True
            self.check_external_variants = False
        else:
            self.process_all_transcripts = False
            self.check_external_variants = True

        self.logger:Logger = None
        self.tally:TallyTable = None

    def load_reference(self):
        """ load reference genome, annotation, and canonical peptides """
        genome, anno, _, canonical_peptides = common.load_references(
                args=self.args,
                invalid_protein_as_noncoding=self.invalid_protein_as_noncoding,
                cleavage_params=self.cleavage_params
            )
        self.reference_data = params.ReferenceData(
            genome=genome,
            anno=anno,
            canonical_peptides=canonical_peptides
        )

    def create_in_disk_variant_pool(self):
        """ Create in disk variant pool """
        self.variant_record_pool = seqvar.VariantRecordPoolOnDisk(
            gvf_files=self.variant_files,
            anno=self.reference_data.anno,
            genome=self.reference_data.genome
        )

    def write_fasta(self):
        """ write variant peptides to fasta. """
        self.variant_peptides.write(self.output_path)
        self.logger.info('Variant peptide FASTA file written to disk.')

    def write_dgraphs(self, tx_id:str, dgraphs:TypeDGraphs):
        """ Write ThreeFrameTVG data """
        if dgraphs[0]:
            data = dgraphs[0].jsonfy()
            with open(self.graph_output_dir/f"{tx_id}_main_TVG.json", 'wt') as handle:
                json.dump(data, handle)

        for var_id, pgraph in dgraphs[1].items():
            if pgraph is None:
                continue
            data = pgraph.jsonfy()
            with open(self.graph_output_dir/f"{tx_id}_Fusion_{var_id}_TVG.json", 'wt') as handle:
                json.dump(data, handle)

        for var_id, pgraph in dgraphs[2].items():
            if pgraph is None:
                continue
            data = pgraph.jsonfy()
            with open(self.graph_output_dir/f"{tx_id}_circRNA_{var_id}_TVG.json", 'wt') as handle:
                json.dump(data, handle)

    def write_pgraphs(self, tx_id:str, pgraphs:TypePGraphs):
        """ Write PeptideVariantGraph data """
        if pgraphs[0]:
            data = pgraphs[0].jsonfy()
            with open(self.graph_output_dir/f"{tx_id}_main_PVG.json", 'wt') as handle:
                json.dump(data, handle)

        for var_id, pgraph in pgraphs[1].items():
            if pgraph is None:
                continue
            data = pgraph.jsonfy()
            with open(self.graph_output_dir/f"{tx_id}_Fusion_{var_id}_PVG.json", 'wt') as handle:
                json.dump(data, handle)

        for var_id, pgraph in pgraphs[2].items():
            if pgraph is None:
                continue
            data = pgraph.jsonfy()
            with open(self.graph_output_dir/f"{tx_id}_circRNA_{var_id}_PVG.json", 'wt') as handle:
                json.dump(data, handle)

TypeDGraphs = Tuple[
    svgraph.ThreeFrameTVG,
    Dict[str, svgraph.ThreeFrameTVG],
    Dict[str, svgraph.ThreeFrameCVG]
]
TypePGraphs = Tuple[
    svgraph.PeptideVariantGraph,
    Dict[str, svgraph.PeptideVariantGraph],
    Dict[str, svgraph.PeptideVariantGraph]
]
# pylint: disable=unused-argument
@common.timeout()
def call_variant_peptides_wrapper(tx_id:str,
        variant_series:seqvar.TranscriptionalVariantSeries,
        tx_seqs:Dict[str, dna.DNASeqRecordWithCoordinates],
        gene_seqs:Dict[str, dna.DNASeqRecordWithCoordinates],
        reference_data:params.ReferenceData,
        pool:seqvar.VariantRecordPool,
        cleavage_params:params.CleavageParams,
        noncanonical_transcripts:bool,
        max_adjacent_as_mnv:bool,
        truncate_sec:bool,
        w2f_reassignment:bool,
        backsplicing_only:bool,
        save_graph:bool,
        coding_novel_orf:bool,
        **kwargs
        ) -> Tuple[Dict[Seq, List[AnnotatedPeptideLabel]], str, TypeDGraphs, TypePGraphs]:
    """ wrapper function to call variant peptides """
    logger = get_logger()
    peptide_anno:Dict[Seq, Dict[str, AnnotatedPeptideLabel]] = {}

    def add_peptide_anno(x:Dict[Seq, List[AnnotatedPeptideLabel]]):
        for seq, seq_data in x.items():
            val = peptide_anno.setdefault(seq, {})
            for metadata in seq_data:
                if metadata.label not in val:
                    val[metadata.label] = metadata

    main_peptides = None
    denylist = call_canonical_peptides(
        tx_id=tx_id, ref=reference_data, tx_seq=tx_seqs[tx_id],
        cleavage_params=cleavage_params, truncate_sec=truncate_sec,
        w2f=w2f_reassignment
    )
    dgraphs:TypeDGraphs = (None, {}, {})
    pgraphs:TypePGraphs = (None, {}, {})
    if variant_series.transcriptional:
        try:
            if not noncanonical_transcripts or \
                    variant_series.has_any_alternative_splicing():
                peptide_map, dgraph, pgraph = call_peptide_main(
                    tx_id=tx_id, tx_variants=variant_series.transcriptional,
                    variant_pool=pool, ref=reference_data,
                    tx_seqs=tx_seqs, gene_seqs=gene_seqs,
                    cleavage_params=cleavage_params,
                    max_adjacent_as_mnv=max_adjacent_as_mnv,
                    truncate_sec=truncate_sec,
                    w2f=w2f_reassignment,
                    denylist=denylist,
                    save_graph=save_graph,
                    coding_novel_orf=coding_novel_orf
                )
                main_peptides = set(peptide_map.keys())
                dgraphs = (dgraph, dgraphs[1], dgraphs[2])
                pgraphs = (pgraph, pgraphs[1], pgraphs[2])
                add_peptide_anno(peptide_map)
        except:
            logger.error('Exception raised from %s', tx_id)
            raise

    exclude_variant_types = ['Fusion', 'Insertion', 'Deletion', 'Substitution', 'circRNA']
    for variant in variant_series.fusion:
        try:
            donor_breakpoint_genomic = reference_data.anno.coordinate_transcript_to_genomic(
                index=variant.location.start - 1, transcript=tx_id
            )
            donor_breakpoint_gene = reference_data.anno.coordinate_genomic_to_gene(
                index=donor_breakpoint_genomic, gene=variant.attrs['GENE_ID']
            ) + 1
            filtered_variants = pool.filter_variants(
                tx_ids=[tx_id], start=0, end=donor_breakpoint_gene,
                exclude_type=exclude_variant_types, intron=False,
                return_coord='transcript'
            )
            variant_pool = copy.copy(pool)
            variant_pool[tx_id] = copy.copy(pool[tx_id])
            variant_pool[tx_id].transcriptional = filtered_variants
            peptide_map, dgraph, pgraph = call_peptide_fusion(
                variant=variant, variant_pool=variant_pool, ref=reference_data,
                tx_seqs=tx_seqs, gene_seqs=gene_seqs, cleavage_params=cleavage_params,
                max_adjacent_as_mnv=max_adjacent_as_mnv, w2f_reassignment=w2f_reassignment,
                denylist=denylist, save_graph=save_graph, coding_novel_orf=coding_novel_orf
            )
            dgraphs[1][variant.id] = dgraph
            pgraphs[1][variant.id] = pgraph
            add_peptide_anno(peptide_map)
        except:
            logger.error(
                "Exception raised from fusion %s, donor: %s, accepter: %s",
                variant.id, tx_id, variant.attrs['ACCEPTER_TRANSCRIPT_ID']
            )
            raise

    if main_peptides:
        denylist.update([str(x) for x in main_peptides])
    for circ_model in variant_series.circ_rna:
        try:
            peptide_map, cgraph, pgraph = call_peptide_circ_rna(
                record=circ_model, variant_pool=pool, gene_seqs=gene_seqs,
                cleavage_params=cleavage_params,
                max_adjacent_as_mnv=max_adjacent_as_mnv,
                w2f_reassignment=w2f_reassignment, denylist=denylist,
                save_graph=save_graph, backsplicing_only=backsplicing_only
            )

        except:
            logger.error("Exception raised from %s", circ_model.id)
            raise
        dgraphs[2][circ_model.id] = cgraph
        pgraphs[2][circ_model.id] = pgraph
        add_peptide_anno(peptide_map)

    peptide_anno = {k:list(v.values()) for k,v in peptide_anno.items()}

    return peptide_anno, tx_id, dgraphs, pgraphs

def caller_reducer(dispatch):
    """ wrapper for ParallelPool. Also reduces the complexity if the run is timed out. """
    max_variants_per_node = dispatch['max_variants_per_node']
    additional_variants_per_misc = dispatch['additional_variants_per_misc']
    tx_id = dispatch['tx_id']
    while True:
        try:
            return call_variant_peptides_wrapper(**dispatch)
        except TimeoutError as e:
            new_dispatch = copy.copy(dispatch)
            p = copy.copy(new_dispatch['cleavage_params'])
            max_variants_per_node = max_variants_per_node[1:]
            if len(max_variants_per_node) == 0:
                max_variants_per_node = (p.max_variants_per_node - 1, )
                if max_variants_per_node[0] <= 0:
                    raise ValueError(f"Failed to finish transcript: {tx_id}") from e
            additional_variants_per_misc = additional_variants_per_misc[1:]
            if len(additional_variants_per_misc) == 0:
                additional_variants_per_misc = (0,)
            p.max_variants_per_node = max_variants_per_node[0]
            p.additional_variants_per_misc = additional_variants_per_misc[0]
            new_dispatch['cleavage_params'] = p
            dispatch = new_dispatch
            get_logger().warning(
                "Transcript %s timed out. Retry with "
                "--max-variants-per-node %i --additional-variants-per-misc %i",
                tx_id, p.max_variants_per_node, p.additional_variants_per_misc
            )

def call_variant_peptide(args:argparse.Namespace) -> None:
    """ Main entry point for calling variant peptide """
    caller = VariantPeptideCaller(args)
    common.print_start_message(args)
    caller.load_reference()
    caller.create_in_disk_variant_pool()
    noncanonical_transcripts = caller.noncanonical_transcripts
    ref = caller.reference_data
    logger = get_logger()
    caller.logger = logger
    caller.tally = TallyTable(logger)

    with ExitStack() as stack:
        pool = stack.enter_context(seqvar.VariantRecordPoolOnDiskOpener(caller.variant_record_pool))
        seq_anno_handle = stack.enter_context(open(caller.peptide_table_output_path, 'w+'))

        peptide_table = svgraph.VariantPeptideTable(seq_anno_handle)
        peptide_table.write_header()

        for file in pool.gvf_files:
            logger.info("Using GVF file: %s", file)
        tx_rank = ref.anno.get_transcript_rank()
        tx_sorted = sorted(pool.pointers.keys(), key=lambda x:tx_rank[x])
        logger.info('Variants sorted')
        caller.tally.n_transcripts_total = len(tx_sorted)

        # Not providing canonical peptide pool to each task for now.
        canonical_peptides = set()

        if caller.threads > 1:
            process_pool = ParallelPool(ncpus=caller.threads)

        logger.info('Start calling variant peptides..')

        dispatches = []
        i = 0
        for tx_id in tx_sorted:
            tx_ids = [tx_id]
            tx_model = ref.anno.transcripts[tx_id]
            variant_series = pool[tx_id]
            if variant_series.is_empty():
                continue
            if noncanonical_transcripts and \
                    not variant_series.has_any_noncanonical_transcripts():
                continue
            tx_ids += variant_series.get_additional_transcripts()

            gene_seqs = {}
            if variant_series.is_gene_sequence_needed():
                gene_id = tx_model.transcript.gene_id
                _chrom = tx_model.transcript.chrom
                gene_model = ref.anno.genes[gene_id]
                gene_seq = gene_model.get_gene_sequence(ref.genome[_chrom])
                gene_seqs[gene_id] = gene_seq

            tx_seqs = {}
            for _tx_id in tx_ids:
                _tx_model = ref.anno.transcripts[_tx_id]
                _chrom = _tx_model.transcript.chrom
                tx_seqs[_tx_id] = _tx_model.get_transcript_sequence(ref.genome[_chrom])
                _gene_id = _tx_model.transcript.gene_id
                if _gene_id not in gene_seqs:
                    _gene_model = ref.anno.genes[_gene_id]
                    _gene_seq = _gene_model.get_gene_sequence(ref.genome[_chrom])
                    gene_seqs[_gene_id] = _gene_seq
            gene_ids = list({ref.anno.transcripts[x].transcript.gene_id for x in tx_ids})

            gene_models = {}
            for gene_id in gene_ids:
                dummy_gene_model = copy.deepcopy(ref.anno.genes[gene_id])
                dummy_gene_model.transcripts = [x for x in dummy_gene_model.transcripts
                    if x in tx_ids]
                gene_models[gene_id] = dummy_gene_model

            dummy_anno = gtf.GenomicAnnotation(
                genes=gene_models,
                transcripts={tx_id:ref.anno.transcripts[tx_id] for tx_id in tx_ids},
                source=ref.anno.source
            )
            reference_data = params.ReferenceData(
                genome=None,
                anno=dummy_anno,
                canonical_peptides=canonical_peptides
            )
            dummy_pool = seqvar.VariantRecordPool()
            dummy_pool.anno = dummy_anno
            dummy_pool.data[tx_id] = variant_series
            for add_tx in tx_ids:
                if add_tx != tx_id:
                    try:
                        dummy_pool[add_tx] = caller.variant_record_pool[add_tx]
                    except KeyError:
                        continue

            dispatch = {
                'tx_id': tx_id,
                'variant_series': variant_series,
                'tx_seqs': tx_seqs,
                'gene_seqs': gene_seqs,
                'reference_data': reference_data,
                'pool': dummy_pool,
                'cleavage_params': caller.cleavage_params,
                'noncanonical_transcripts': noncanonical_transcripts,
                'max_adjacent_as_mnv': caller.max_adjacent_as_mnv,
                'truncate_sec': caller.truncate_sec,
                'w2f_reassignment': caller.w2f_reassignment,
                'save_graph': caller.graph_output_dir is not None,
                'timeout': args.timeout_seconds,
                'max_variants_per_node': tuple(args.max_variants_per_node),
                'additional_variants_per_misc': tuple(args.additional_variants_per_misc),
                'backsplicing_only': args.backsplicing_only,
                'coding_novel_orf': args.coding_novel_orf
            }
            dispatches.append(dispatch)

            caller.tally.n_transcripts_processed += 1

            reloaded = ((i + 1) % caller.threads == 0 or i + 1 == len(tx_sorted)) \
                and len(dispatches) > 0
            if reloaded:
                logger.debug([x['tx_id'] for x in dispatches])
                if caller.threads > 1:
                    results = process_pool.map(
                        caller_reducer,
                        dispatches
                    )
                else:
                    results = [ caller_reducer(dispatches[0]) ]

                # pylint: disable=W0621
                for peptide_anno, tx_id, dgraphs, pgraphs in results:
                    caller.tally.n_total_peptides += len(peptide_anno)
                    for peptide in peptide_anno:
                        is_valid = peptide_table.is_valid(
                            seq=peptide,
                            canonical_peptides=ref.canonical_peptides,
                            cleavage_params=caller.cleavage_params
                        )
                        if is_valid:
                            for seq_anno in peptide_anno[peptide]:
                                peptide_table.add_peptide(peptide, seq_anno)
                    if caller.graph_output_dir is not None:
                        caller.write_dgraphs(tx_id, dgraphs)
                        caller.write_pgraphs(tx_id, pgraphs)
                dispatches = []

            i += 1
            if i % 1000 == 0:
                logger.info(
                    '%.2f %% ( %i / %i ) transcripts processed.',
                    i/len(tx_sorted) * 100, i, len(tx_sorted)
                )
        caller.tally.n_valid_peptides = len(peptide_table.index)
        logger.info('All transcripts processed. Preparing FASTA file..')
        peptide_table.write_fasta(caller.output_path)
        logger.info('Variant peptide FASTA written.')
        caller.tally.log()

def call_canonical_peptides(tx_id:str, ref:params.ReferenceData,
        tx_seq:dna.DNASeqRecordWithCoordinates,
        cleavage_params:params.CleavageParams,
        truncate_sec:bool, w2f:bool):
    """ Call canonical peptides """
    tx_model = ref.anno.transcripts[tx_id]
    dgraph = svgraph.ThreeFrameTVG(
        seq=tx_seq, _id=tx_id,
        has_known_orf=tx_model.is_protein_coding,
        mrna_end_nf=False,
        cleavage_params=cleavage_params,
        coordinate_feature_type='transcript',
        coordinate_feature_id=tx_id
    )
    dgraph.gather_sect_variants(ref.anno)
    dgraph.init_three_frames()
    pgraph = dgraph.translate()
    pgraph.create_cleavage_graph()
    peptide_map = pgraph.call_variant_peptides(
        check_variants=False, truncate_sec=truncate_sec, w2f=w2f,
        check_external_variants=False
    )
    return {str(x) for x in peptide_map}

if TYPE_CHECKING:
    TypeCallPeptideReturnData = Tuple[
        Dict[Seq, List[AnnotatedPeptideLabel]],
        Union[ThreeFrameTVG, ThreeFrameCVG],
        PeptideVariantGraph
    ]

def call_peptide_main(tx_id:str, tx_variants:List[seqvar.VariantRecord],
        variant_pool:seqvar.VariantRecordPoolOnDisk,
        ref:params.ReferenceData,
        tx_seqs:Dict[str, dna.DNASeqRecordWithCoordinates],
        gene_seqs:Dict[str, dna.DNASeqRecordWithCoordinates],
        cleavage_params:params.CleavageParams,
        max_adjacent_as_mnv:bool, truncate_sec:bool, w2f:bool,
        denylist:Set[str], save_graph:bool, coding_novel_orf:bool
        ) -> TypeCallPeptideReturnData:
    """ Call variant peptides for main variants (except cirRNA). """
    tx_model = ref.anno.transcripts[tx_id]
    tx_seq = tx_seqs[tx_id]

    dgraph = svgraph.ThreeFrameTVG(
        seq=tx_seq,
        _id=tx_id,
        cds_start_nf=tx_model.is_cds_start_nf(),
        has_known_orf=tx_model.is_protein_coding,
        mrna_end_nf=tx_model.is_mrna_end_nf(),
        cleavage_params=cleavage_params,
        max_adjacent_as_mnv=max_adjacent_as_mnv,
        coordinate_feature_type='transcript',
        coordinate_feature_id=tx_id
    )
    dgraph.gather_sect_variants(ref.anno)
    dgraph.init_three_frames()
    dgraph.create_variant_graph(
        variants=tx_variants, variant_pool=variant_pool, genome=ref.genome,
        anno=ref.anno, tx_seqs=tx_seqs, gene_seqs=gene_seqs
    )
    dgraph.fit_into_codons()
    pgraph = dgraph.translate()

    pgraph.create_cleavage_graph()

    if tx_model.is_protein_coding:
        peptide_map = pgraph.call_variant_peptides(
            denylist=denylist,
            truncate_sec=truncate_sec,
            w2f=w2f,
            check_external_variants=True,
            check_orf=False
        )
    else:
        peptide_map = {}

    if not tx_model.is_protein_coding or coding_novel_orf:
        peptide_novel_orf = pgraph.call_variant_peptides(
            denylist=denylist,
            truncate_sec=truncate_sec,
            w2f=w2f,
            check_external_variants=True,
            check_orf=True
        )
        for peptide, labels in peptide_novel_orf.items():
            if peptide not in peptide_map:
                peptide_map[peptide] = labels

    if not save_graph:
        dgraph, pgraph = None, None
    return peptide_map, dgraph, pgraph

def call_peptide_fusion(variant:seqvar.VariantRecord,
        variant_pool:seqvar.VariantRecordPoolOnDisk,
        ref:params.ReferenceData,
        tx_seqs:Dict[str, dna.DNASeqRecordWithCoordinates],
        gene_seqs:Dict[str, dna.DNASeqRecordWithCoordinates],
        cleavage_params:params.CleavageParams, max_adjacent_as_mnv:bool,
        w2f_reassignment:bool, denylist:Set[str], save_graph:bool,
        coding_novel_orf:bool
        ) -> TypeCallPeptideReturnData:
    """ Call variant peptides for fusion """
    tx_id = variant.location.seqname
    tx_model = ref.anno.transcripts[tx_id]
    tx_seq = tx_seqs[tx_id]
    tx_seq = tx_seq[:variant.location.start]
    donor_tx_id = variant.accepter_transcript_id

    orf_start = tx_seq.orf.start + 3 if tx_seq.orf else 3
    if variant.location.start < orf_start:
        return {}, None, None

    if tx_id in variant_pool:
        tx_variants = [x for x in variant_pool[tx_id].transcriptional
            if x.location.end < variant.location.start]
    else:
        tx_variants = []

    tx_variants.append(variant)

    dgraph = svgraph.ThreeFrameTVG(
        seq=tx_seq,
        _id=tx_id,
        cds_start_nf=tx_model.is_cds_start_nf(),
        has_known_orf=tx_model.is_protein_coding,
        mrna_end_nf=ref.anno.transcripts[donor_tx_id].is_mrna_end_nf(),
        cleavage_params=cleavage_params,
        max_adjacent_as_mnv=max_adjacent_as_mnv,
        coordinate_feature_type='transcript',
        coordinate_feature_id=tx_id
    )
    dgraph.gather_sect_variants(ref.anno)
    dgraph.sect_variants = [v for v in dgraph.sect_variants
        if v.location.end < variant.location.start]
    dgraph.init_three_frames()
    dgraph.create_variant_graph(
        variants=tx_variants, variant_pool=variant_pool, genome=ref.genome,
        anno=ref.anno, tx_seqs=tx_seqs, gene_seqs=gene_seqs
    )
    dgraph.fit_into_codons()
    pgraph = dgraph.translate()
    pgraph.create_cleavage_graph()

    if tx_model.is_protein_coding:
        peptide_map = pgraph.call_variant_peptides(
            denylist=denylist,
            w2f=w2f_reassignment,
            check_external_variants=True,
            check_orf=False
        )
    else:
        peptide_map = {}

    if not tx_model.is_protein_coding or coding_novel_orf:
        peptide_novel_orf = pgraph.call_variant_peptides(
            denylist=denylist,
            w2f=w2f_reassignment,
            check_external_variants=True,
            check_orf=True
        )
        for peptide, labels in peptide_novel_orf.items():
            if peptide not in peptide_map:
                peptide_map[peptide] = labels

    if not save_graph:
        dgraph, pgraph = None, None
    return peptide_map, dgraph, pgraph

def call_peptide_circ_rna(record:circ.CircRNAModel,
        variant_pool:seqvar.VariantRecordPool,
        gene_seqs:Dict[str, dna.DNASeqRecordWithCoordinates],
        cleavage_params:params.CleavageParams, max_adjacent_as_mnv:bool,
        backsplicing_only:bool, w2f_reassignment:bool, denylist:Set[str],
        save_graph:bool
        )-> TypeCallPeptideReturnData:
    """ Call variant peptides from a given circRNA """
    gene_id = record.gene_id
    gene_seq = gene_seqs[gene_id]
    record.fragments.sort()
    circ_seq = record.get_circ_rna_sequence(gene_seq)
    for loc in circ_seq.locations:
        loc.ref.seqname = record.id

    # Alternative splicing should not be included. Alternative splicing
    # represented as Insertion, Deletion or Substitution.
    exclusion_variant_types = ['Insertion', 'Deletion', 'Substitution']

    fragments = []
    for frag in record.fragments:
        if len(frag) <= 3:
            continue
        loc = FeatureLocation(
            start=frag.location.start + 3, end=frag.location.end
        )
        frag = SeqFeature(chrom=frag.chrom, location=loc, attributes=frag.attributes)
        fragments.append(frag)

    if not fragments:
        return {}, None, None

    variant_records = variant_pool.filter_variants(
        tx_ids=[record.transcript_id], exclude_type=exclusion_variant_types,
        intron=False, segments=fragments
    )

    cgraph = svgraph.ThreeFrameCVG(
        circ_seq, _id=record.id, circ_record=record,
        cleavage_params=cleavage_params,
        max_adjacent_as_mnv=max_adjacent_as_mnv,
        coordinate_feature_type='gene',
        coordinate_feature_id=gene_id
    )
    cgraph.init_three_frames()
    cgraph.create_variant_circ_graph(variant_records)
    cgraph.extend_loop()
    cgraph.truncate_three_frames()
    cgraph.fit_into_codons()
    pgraph = cgraph.translate()
    pgraph.create_cleavage_graph()
    peptide_map = pgraph.call_variant_peptides(
        denylist=denylist, circ_rna=record, backsplicing_only=backsplicing_only,
        w2f=w2f_reassignment, check_external_variants=True
    )
    if not save_graph:
        cgraph, pgraph = None, None
    return peptide_map, cgraph, pgraph



r""" Downsample reference FASTA and GTF

command line usage:
    python -m test.downsample_reference --help

    python -m test.downsample_reference \
        --genome-fasta path/to/genome.fasta \
        --annotation-gtf path/to/annotation.gtf \
        --protein-fasta path/to/translate.fasta \
        --tx-list ENST0001 ENST0002 \
        --output-dir path/to/downsampled_index

    python -m test.downsample_reference \
        --genome-fasta path/to/genome.fasta \
        --annotation-gtf path/to/annotation.gtf \
        --protein-fasta path/to/translate.fasta \
        --gene-list ENSG0001 ENSG0002 \
        --output-dir path/to/downsampled_index
"""
import argparse
import copy
from typing import List, Tuple, Iterable, Dict
from pathlib import Path
import math
from Bio import SeqIO
from moPepGen import gtf, dna, aa
from moPepGen.gtf import GtfIO
from moPepGen.SeqFeature import FeatureLocation, SeqFeature
from moPepGen.gtf.GTFSeqFeature import GTFSeqFeature
from moPepGen.cli.common import add_args_cleavage, add_args_reference, \
    print_help_if_missing_args, add_args_debug_level


# pylint: disable=W0212
def parse_args(subparsers:argparse._SubParsersAction):
    """ Parse args """
    parser:argparse.ArgumentParser = subparsers.add_parser(
        name='downsampleReference',
        help='Create downsampleed reference files to contain only specified'
        ' transcripts.'
    )
    parser.add_argument(
        '--gene-list',
        type=str,
        help='Gene list to filter.',
        nargs='*',
        metavar='<value>'
    )
    parser.add_argument(
        '--tx-list',
        type=str,
        help='Transcript list to filter.',
        nargs='*',
        metavar='<value>'
    )
    parser.add_argument(
        '--output-dir',
        type=Path,
        help='Output directory',
        metavar='<file>'
    )
    parser.add_argument(
        '--translate-noncoding',
        type=str,
        choices=['true', 'false'],
        help='Create translate sequences for noncoding at any possible ORF.',
        default='true'
    )
    add_args_reference(parser, index=False)
    add_args_cleavage(parser)
    add_args_debug_level(parser)
    parser.set_defaults(func=main)
    print_help_if_missing_args(parser)
    return parser

GeneTranscriptModel = Tuple[gtf.GeneAnnotationModel, Dict[str, \
    gtf.TranscriptAnnotationModel]]

def parse_gtf(path:Path) -> Iterable[GeneTranscriptModel]:
    """ Parse GTF """
    it = GtfIO.parse(path)
    gene:gtf.GeneAnnotationModel = None
    transcripts:Dict[str, gtf.TranscriptAnnotationModel] = {}

    while True:
        record:GTFSeqFeature = next(it, None)
        if not record:
            if gene:
                yield gene, transcripts
            return
        if record.type.lower() == 'gene':
            if gene:
                yield gene, transcripts
            gene = record
            gene.__class__ = gtf.GeneAnnotationModel
            gene.transcripts = []
            transcripts = {}
            continue
        feature = record.type.lower()
        if feature not in gtf.GTF_FEATURE_TYPES:
            continue
        transcript_id = record.transcript_id
        if transcript_id not in transcripts:
            transcripts[transcript_id] = gtf.TranscriptAnnotationModel()
            gene.transcripts.append(transcript_id)
        transcripts[transcript_id].add_record(feature, record)

def downsample_gtf(path:Path, gene_list=None, tx_list=None) -> gtf.GeneAnnotationModel:
    """ Downsample a GTF file """
    anno = gtf.GenomicAnnotation()

    for model in parse_gtf(path):
        if gene_list:
            gene_id = model[0].gene_id
            if gene_id in gene_list:
                anno.genes[gene_id] = model[0]
                anno.transcripts.update(model[1])
        if tx_list:
            gene_wanted = False
            tx_ids = []
            for key,val in model[1].items():
                if key in tx_list:
                    gene_wanted = True
                    anno.transcripts[key] = val
                    tx_ids.append(key)
            if gene_wanted:
                gene_id = model[0].gene_id
                anno.genes[gene_id] = model[0]
                anno.genes[gene_id].transcripts = tx_ids
    for transcript in anno.transcripts.values():
        transcript.sort_records()
    return anno


def get_gene_sequences(path:Path, anno:gtf.GenomicAnnotation) -> dna.DNASeqDict:
    """ Get gene sequences """
    gene_seqs = dna.DNASeqDict()
    chrom_genes:Dict[str, List[gtf.GeneAnnotationModel]] = {}
    for gene_model in anno.genes.values():
        chrom = gene_model.location.seqname
        if chrom not in chrom_genes:
            chrom_genes[chrom] = []
        chrom_genes[chrom].append(gene_model)

    for record in SeqIO.parse(path, 'fasta'):
        record.__class__ = dna.DNASeqRecord
        record.id = record.id.split(' ')[0]
        record.id = record.id.split('|')[0]
        if record.id not in chrom_genes:
            continue
        for gene_model in chrom_genes[record.id]:
            gene_id = gene_model.gene_id
            seq = gene_model.get_gene_sequence(record)
            seq.__class__ = dna.DNASeqRecord
            gene_seqs[gene_id] = seq
    return gene_seqs

def downsample_proteins(path:Path, anno:gtf.GenomicAnnotation
        ) -> aa.AminoAcidSeqDict:
    """ Downsample proteins """
    proteins = aa.AminoAcidSeqDict()
    tx_ids = list(anno.transcripts.keys())

    count = 0
    source = None
    infered = set()

    for record in SeqIO.parse(path, 'fasta'):
        record.__class__ = aa.AminoAcidSeqRecord
        record:aa.AminoAcidSeqRecord
        if count > 100 and not source:
            source = infered.pop()
        if not source:
            count += 1
            infered.add(record.infer_ids())
        else:
            record.infer_ids(style=source)

        if record.transcript_id in tx_ids:
            proteins[record.transcript_id] = record
    return proteins

def shift_seq_feature(feature:SeqFeature, offset:int, seqname:str='chr1'):
    """ Shift SeqFeature with a given offset """
    feature.chrom = seqname
    start = feature.location.start + offset
    end = feature.location.end + offset
    strand = feature.location.strand
    location = FeatureLocation(seqname=seqname, start=start, end=end, strand=strand)
    feature.location = location

def shift_reference(gene_seqs:dna.DNASeqDict, anno:gtf.GenomicAnnotation
        ) -> Tuple[dna.DNASeqDict, gtf.GenomicAnnotation]:
    """ Shift reference """
    seqname = 'chr1'
    genome = dna.DNASeqDict({seqname: None})
    for gene_id, gene_seq in gene_seqs.items():
        gene_seq:dna.DNASeqRecord
        strand = anno.genes[gene_id].location.strand
        seq = gene_seq
        if strand == -1:
            seq = seq.reverse_complement()
        if not genome[seqname]:
            genome[seqname] = seq
        else:
            genome[seqname] += seq
    genome[seqname].id = seqname

    gene_start = 0
    for gene_id in gene_seqs.keys():
        gene = anno.genes[gene_id]
        shift_offset = gene_start - gene.location.start
        gene_start = gene_start + gene.location.end - gene.location.start
        shift_seq_feature(gene, shift_offset, seqname)

        for tx_id in gene.transcripts:
            tx_model = anno.transcripts[tx_id]
            shift_seq_feature(tx_model.transcript, shift_offset, seqname)
            for exon in tx_model.exon:
                shift_seq_feature(exon, shift_offset, seqname)
            for cds in tx_model.cds:
                shift_seq_feature(cds, shift_offset, seqname)
            for utr in tx_model.utr:
                shift_seq_feature(utr, shift_offset, seqname)
            for sec in tx_model.selenocysteine:
                shift_seq_feature(sec, shift_offset, seqname)
    return genome, anno

def get_noncoding_translate(tx_id:str, anno:gtf.GenomicAnnotation,
        genome:dna.DNASeqDict) -> Dict[str, aa.AminoAcidSeqRecord]:
    """ Translate all possible ORF of a noncoding transcript """
    tx_model = anno.transcripts[tx_id]
    protein_id = tx_id.replace('ENST', 'ENSP')
    gene_id = tx_model.transcript.gene_id
    chrom = tx_model.transcript.chrom
    tx_seq = tx_model.get_transcript_sequence(genome[chrom])
    start_positions = tx_seq.find_all_start_codons()

    translates = {}
    for start in start_positions:
        end = start + math.floor((len(tx_seq) - start) / 3) * 3
        aa_seq = tx_seq[start:end].translate(to_stop=True)
        end = start + len(aa_seq) * 3
        orf = f"ORF{start}:{end}"
        alt_protein_id = f"{protein_id}-{orf}"
        alt_tx_id = f"{tx_id}-{orf}"
        description = f"{alt_protein_id}|{alt_tx_id}|{gene_id}|-"
        aa_seq.id = alt_protein_id
        aa_seq.protein_id = alt_protein_id
        aa_seq.transcript_id = alt_tx_id
        aa_seq.gene_id = gene_id
        aa_seq.description = description
        translates[alt_tx_id] = aa_seq
    return translates

def create_dummy_tx_model(tx_model:gtf.TranscriptAnnotationModel, tx_id:str,
        protein_id:str):
    """ Create dummy transcript model """
    location = FeatureLocation(
        start=tx_model.transcript.location.start,
        end=tx_model.transcript.location.end,
        strand=tx_model.transcript.location.strand
    )
    attrs = copy.deepcopy(tx_model.transcript.attributes)
    attrs['protein_id'] = protein_id
    attrs['transcript_id'] = tx_id
    transcript = GTFSeqFeature(
        location=location, chrom=tx_model.transcript.chrom,
        attributes=attrs, type='transcript'
    )
    model = gtf.TranscriptAnnotationModel(transcript)
    return model

def subset_genome(genome:dna.DNASeqDict, anno:gtf.GenomicAnnotation,
        start:int, end:int) -> Tuple[dna.DNASeqDict, gtf.GenomicAnnotation]:
    """ Subset genome """
    if anno.genes > 1:
        raise ValueError(
            "Don't know how to subset the genome when there are multiple genes."
        )
    gene_id = list(anno.genes.keys())[0]
    gene_model = anno.genes[gene_id]
    if start > gene_model.location.end or end <= gene_model.location.start:
        raise ValueError("Subset location is invalid.")

    for tx_id in list(anno.transcripts.keys()):
        tx_model = anno.transcripts[tx_id]
        if tx_model.transcript.location.end < start:
            anno.transcripts.pop(tx_id)
            continue
        original_start = tx_model.transcript.location.start
        original_end = tx_model.transcript.location.end
        tx_start = max(original_start - start, 0)
        tx_end = min(original_end - start, end - start)
        tx_model.transcript.location = FeatureLocation(
            start=tx_start, end=tx_end, seqname=tx_model.transcript.seqname
        )

        tx_model.exon = subset_and_filter(tx_model.exon, start, end)
        tx_model.cds = subset_and_filter(tx_model.cds, start, end, True)
        tx_model.utr = subset_and_filter(tx_model.utr, start, end)

    seqname = list(genome.keys())[0]
    if len(genome[seqname]) < start or len(genome[seqname]) < end:
        raise ValueError("Don't know how to subset genome like this.")

    genome[seqname] = genome[seqname][start:end]

def subset_and_filter(features:List[SeqFeature], start:int, end:int,
        cds:bool=False) -> List[SeqFeature]:
    """ Subset and filter SeqFeatures """
    new_features = []
    for i, feature in enumerate(features):
        original_start = feature.location.start
        original_end = feature.location.end
        if original_start > end or start > original_end:
            continue
        new_start = max(original_start - start, 0)
        new_end = min(original_end - start, end - start)
        if cds:
            if i == 0 and feature.strand == 1:
                if new_start == 0:
                    new_start += (3 - (start - original_start) % 3) % 3
            if i == len(features) and feature.strand == -1:
                if new_end == end - start:
                    new_end -= (3 - (original_end - end) % 3) % 3
        location = FeatureLocation(
            start=new_start, end=new_end, seqname=feature.seqname,
            strand=feature.strand
        )
        new_feature = SeqFeature(
            chrom=feature.chrom,
            attributes=feature.attributes,
            location=location
        )
        new_features.append(new_feature)
    return new_features


def main(args:argparse.Namespace):
    """ Downsample reference FASTA and GTF """
    genome_fasta:Path = args.genome_fasta
    annotation_gtf:Path = args.annotation_gtf
    protein_fasta:Path = args.proteome_fasta
    gene_list:List[str] = args.gene_list
    tx_list:List[str] = args.tx_list
    output_dir:Path = args.output_dir
    translate_noncoding:bool = args.translate_noncoding == 'true'

    output_dir.mkdir(exist_ok=True)

    anno = downsample_gtf(annotation_gtf, gene_list, tx_list)
    gene_seqs = get_gene_sequences(genome_fasta, anno)
    genome, anno = shift_reference(gene_seqs, anno)
    proteins = downsample_proteins(protein_fasta, anno)

    if translate_noncoding:
        for tx_id in copy.copy(list(anno.transcripts.keys())):
            if tx_id not in proteins:
                aa_seqs = get_noncoding_translate(tx_id, anno, genome)
                proteins.update(aa_seqs)

    with open(output_dir/'annotation.gtf', 'wt') as handle:
        GtfIO.write(handle, anno)

    with open(output_dir/'genome.fasta', 'wt') as handle:
        writer = SeqIO.FastaIO.FastaWriter(handle, record2title=lambda x: x.id)
        for record in genome.values():
            writer.write_record(record)

    with open(output_dir/'proteome.fasta', 'wt') as handle:
        writer = SeqIO.FastaIO.FastaWriter(handle, record2title=lambda x: x.description)
        for record in proteins.values():
            writer.write_record(record)



""" Module for rMATS parser """
from __future__ import annotations
from typing import List
from abc import ABC, abstractmethod
from moPepGen import seqvar

class RMATSRecord(ABC):
    """ rMATS record """
    def __init__(self, gene_id:str, gene_symbol:str, chrom:str):
        """"""
        self.gene_id = gene_id
        self.gene_symbol = gene_symbol
        self.chrom = chrom

    @abstractmethod
    def convert_to_variant_records(self, anno, genome, min_ijc, min_sjc
            ) -> List[seqvar.VariantRecord]:
        """ Convert to variant records """



""" Module to collapse TNGNode if they have the same sequence """
from __future__ import annotations
from typing import Dict, Set
from moPepGen import get_equivalent
from moPepGen.svgraph.TVGEdge import TVGEdge
from moPepGen.svgraph.TVGNode import TVGNode


class TVGCollapseNode(TVGNode):
    """ Node for collapsing """
    @classmethod
    def from_tvg_node(cls, node:TVGNode) -> TVGCollapseNode:
        """ Convert from a PVGNode """
        collapse_node = node.copy()
        collapse_node.__class__ = cls
        for edge in node.in_edges:
            edge_copy = TVGEdge(edge.in_node, collapse_node, edge.type)
            collapse_node.in_edges.add(edge_copy)
        for edge in node.out_edges:
            edge_copy = TVGEdge(collapse_node, edge.out_node, edge.type)
            collapse_node.out_edges.add(edge_copy)
        return collapse_node

    def __eq__(self, other:TVGCollapseNode):
        """ equal to """
        result = self.seq.seq == other.seq.seq \
            and set(self.get_in_nodes()) == set(other.get_in_nodes()) \
            and set(self.get_out_nodes()) == set(other.get_out_nodes()) \
            and self.reading_frame_index == other.reading_frame_index \
            and self.subgraph_id == other.subgraph_id \
            and self.was_bridge == other.was_bridge

        if result and hasattr(other, 'match'):
            other.match = self
        return result

    def __ne__(self, other:TVGCollapseNode):
        """ not equal to """
        return not self == other

    def __hash__(self):
        """ hash """
        hash_values = (
            self.seq.seq, frozenset(self.get_out_nodes()),
            frozenset(self.get_in_nodes()), self.reading_frame_index,
            self.subgraph_id, self.was_bridge
        )
        return hash(hash_values)

class TVGNodeCollapser():
    """ Collapse PVGNode """
    def __init__(self, pool:Set[TVGNode]=None, mapper:Dict[TVGNode, TVGNode]=None):
        """ constructor """
        self.pool = pool or set()
        self.mapper = mapper or {}

    @staticmethod
    def first_is_stop_altering(first:TVGNode, second:TVGNode) -> bool:
        """ Check whether the first node is stop altering while the second
        is not. """
        return any(x.is_stop_altering for x in first.variants) \
            and not any(y.is_stop_altering for y in second.variants)

    def should_keep_first(self, first:TVGNode, second:TVGNode) -> bool:
        """ Here we keep the node with least number of variants, unless one
        has stop altering mutation. """
        if self.first_is_stop_altering(first=first, second=second):
            return True
        if self.first_is_stop_altering(first=second, second=first):
            return False
        if first.is_less_mutated_than(second):
            return True
        return False

    def collapse(self, node:TVGNode) -> TVGNode:
        """ Collapse the given node if it has the same in the pool """
        collapse_node = TVGCollapseNode.from_tvg_node(node)
        same_collapse_node = get_equivalent(self.pool, collapse_node)
        if same_collapse_node:
            same_node = self.mapper[same_collapse_node]

            if self.should_keep_first(node, same_node):
                self.pool.remove(same_collapse_node)
                self.pool.add(collapse_node)
                self.mapper.pop(same_collapse_node, None)
                self.mapper[collapse_node] = node
                node_to_discard = same_node
            else:
                node_to_discard = node
        else:
            self.pool.add(collapse_node)
            self.mapper[collapse_node] = node
            node_to_discard = None

        return node_to_discard



""" Module for CIRCexplorer parser """
from __future__ import annotations
from pathlib import Path
from typing import Iterable, List, Tuple, Union
from moPepGen.SeqFeature import FeatureLocation, SeqFeature
from moPepGen import gtf
from moPepGen.circ import CircRNAModel


class CIRCexplorer2KnownRecord():
    """ CIRCexplorer Known Record
    NOTE: CIRCexplorer uses 0-based and end exclusive coordinates.
    """
    def __init__(self, chrom:str, start:int, end:int, name:str, score:float,
            strand:str, thick_start:int, thick_end:int,
            item_rgb:Tuple[int,int,int], exon_count:int, exon_sizes:List[int],
            exon_offsets:List[int], read_number: int, circ_type:str,
            gene_name:str, isoform_name:str, index:List[int], flank_intron:str):
        """ Constructor """
        self.chrom = chrom
        self.start = start
        self.end = end
        self.name = name
        self.score = score
        self.strand = strand
        self.thick_start = thick_start
        self.thick_end = thick_end
        self.item_rgb = item_rgb
        self.exon_count = exon_count
        self.exon_sizes = exon_sizes
        self.exon_offsets = exon_offsets
        self.read_number = read_number
        self.circ_type = circ_type
        self.gene_name = gene_name
        self.isoform_name = isoform_name
        self.index = index
        self.flank_intron = flank_intron

    def convert_to_circ_rna(self, anno:gtf.GenomicAnnotation,
            intron_start_range:Tuple[int,int]=(0,0),
            intron_end_range:Tuple[int,int]=(0,0)) -> CircRNAModel:
        """ Convert a CIRCexplorerKnownRecord to CircRNAModel. """
        tx_id = self.isoform_name
        tx_model = anno.transcripts[tx_id]
        gene_id = tx_model.transcript.gene_id
        strand = tx_model.transcript.strand

        fragments:List[SeqFeature] = []
        intron:List[int] = []

        if self.circ_type == 'circRNA':
            fragment_type = 'exon'
            circ_id = 'CIRC'
        elif self.circ_type == 'ciRNA':
            fragment_type = 'intron'
            circ_id = 'CI'
        else:
            raise ValueError(f'circRNA type unsupported: {self.circ_type}')

        fragment_ids:List[Tuple(fragment, str, int)] = []

        for i, exon_size in enumerate(self.exon_sizes):
            exon_offset = self.exon_offsets[i]
            start = anno.coordinate_genomic_to_gene(
                self.start + exon_offset, gene_id)
            end = anno.coordinate_genomic_to_gene(
                self.start + exon_offset + exon_size - 1, gene_id)

            if strand == -1:
                start, end = end, start
            end += 1

            location = FeatureLocation(
                seqname=gene_id, start=start, end=end, strand=strand
            )

            fragment = SeqFeature(
                chrom=tx_id, location=location, attributes={},
                type=fragment_type
            )

            if fragment_type == 'exon':
                exon_index = anno.find_exon_index(tx_id, fragment)
                fragment_ids.append((fragment, 'E', exon_index))

            elif fragment_type == 'intron':
                intron_index = anno.find_intron_index(
                    tx_id, fragment,
                    intron_start_range=intron_start_range,
                    intron_end_range=intron_end_range
                )
                fragment_ids.append((fragment, 'I', intron_index))
                intron.append(i)

            fragments.append(fragment)

        fragment_ids.sort(key=lambda x: x[0])
        circ_id += f'-{tx_id}-' + '-'.join([f"{t}{i+1}" for _,t,i in fragment_ids])

        genomic_location = f"{self.chrom}:{self.start}"

        return CircRNAModel(
            transcript_id=tx_id,
            fragments=fragments,
            intron=intron,
            _id=circ_id,
            gene_id=gene_id,
            gene_name=tx_model.transcript.gene_name,
            genomic_location=genomic_location
        )

    def is_valid(self, min_read_number:int) -> bool:
        """ Check if this is valid CIRCexplorer record """
        return self.read_number >= min_read_number

class CIRCexplorer3KnownRecord(CIRCexplorer2KnownRecord):
    """ CIRCexplorer3 """
    def __init__(self, *args, fpb_circ:float, fpb_linear:float,
            circ_score:float, **kwargs):
        super().__init__(*args, **kwargs)
        self.fpb_circ = fpb_circ
        self.fpb_linear = fpb_linear
        self.circ_score = circ_score

    def is_valid(self, min_read_number: int, min_fbr_circ:float=None,
            min_circ_score:float=None) -> bool:
        """ Check if is valid """
        #pylint: disable=W0221
        if min_fbr_circ and self.fpb_circ < min_fbr_circ:
            return False
        if min_circ_score and self.circ_score < min_circ_score:
            return False
        return super().is_valid(min_read_number)


CIRCexplorerKnownRecord = Union[CIRCexplorer2KnownRecord, CIRCexplorer3KnownRecord]

def parse(path:Path, circ_explorer_3:bool=False
        ) -> Iterable[CIRCexplorerKnownRecord]:
    """ parse """
    with open(path, 'rt') as handle:
        for line in handle:
            fields = line.rstrip().split('\t')
            data = {
                'chrom': fields[0],
                'start': int(fields[1]),
                'end': int(fields[2]),
                'name': fields[3],
                'score': float(fields[4]),
                'strand': fields[5],
                'thick_start': int(fields[6]),
                'thick_end': int(fields[7]),
                'item_rgb': [int(x) for x in fields[8].split(',')],
                'exon_count': int(fields[9]),
                'exon_sizes': [int(x) for x in fields[10].split(',')],
                'exon_offsets': [int(x) for x in fields[11].split(',')],
                'read_number': int(fields[12]),
                'circ_type': fields[13],
                'gene_name': fields[14],
                'isoform_name': fields[15],
                'index': [int(x) for x in fields[16].split(',')],
                'flank_intron': fields[17]
            }
            if circ_explorer_3:
                yield CIRCexplorer3KnownRecord(
                    fpb_circ=float(fields[18]),
                    fpb_linear=float(fields[19]),
                    circ_score=float(fields[20]),
                    **data
                )
            else:
                yield CIRCexplorer2KnownRecord(**data)



""" `callAltTranslation` calls peptide sequences from coding transcripts that
harbor any alternative translation event. """
from __future__ import annotations
import argparse
from typing import TYPE_CHECKING, Dict, Set
from pathlib import Path
from moPepGen import params, svgraph, aa, get_logger, VARIANT_PEPTIDE_SOURCE_DELIMITER
from moPepGen.cli import common


if TYPE_CHECKING:
    from Bio.Seq import Seq
    from moPepGen.gtf import TranscriptAnnotationModel, GenomicAnnotation
    from moPepGen.dna import DNASeqDict

OUTPUT_FILE_FORMATS = ['.fa', '.fasta']

# pylint: disable=W0212
def add_subparser_call_alt_translation(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen callAltTranslation """
    p:argparse.ArgumentParser = subparsers.add_parser(
        name='callAltTranslation',
        help='Call peptides with alternative translation from coding transcripts.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    p.add_argument(
        '-o', '--output-path',
        type=Path,
        help='Output path to the alternative translation peptide FASTA.'
        f" Valid formats: {OUTPUT_FILE_FORMATS}",
        metavar='<file>',
        required=True
    )
    p.add_argument(
        '--w2f-reassignment',
        action='store_true',
        help='Include peptides with W > F (Tryptophan to Phenylalanine) '
        'reassignment.'
    )
    p.add_argument(
        '--selenocysteine-termination',
        action='store_true',
        help='Include peptides of selenoprotiens that the UGA is treated as '
        'termination instead of Sec.'
    )

    common.add_args_reference(p)
    common.add_args_cleavage(p)
    common.add_args_debug_level(p)

    p.set_defaults(func=call_alt_translation)
    common.print_help_if_missing_args(p)
    return p


def call_alt_translation(args:argparse.Namespace) -> None:
    """ Main entrypoint for calling alternative translation peptides """
    logger = get_logger()

    common.validate_file_format(
        args.output_path, OUTPUT_FILE_FORMATS, check_writable=True
    )

    cleavage_params = params.CleavageParams(
        enzyme=args.cleavage_rule,
        exception=args.cleavage_exception,
        miscleavage=int(args.miscleavage),
        min_mw=float(args.min_mw),
        min_length=args.min_length,
        max_length=args.max_length
    )

    if not (args.selenocysteine_termination or args.w2f_reassignment):
        raise ValueError(
            'At least one of --selenocysteine-termination and --w2f-reassignment'
            ' must be given.'
        )

    common.print_start_message(args)

    genome, anno, _, canonical_peptides = common.load_references(
        args=args, load_proteome=True, cleavage_params=cleavage_params
    )

    peptide_pool = aa.VariantPeptidePool()

    for tx_id in anno.transcripts:
        tx_model = anno.transcripts[tx_id]
        if not tx_model.is_protein_coding:
            continue

        try:
            peptides = call_alt_translation_main(
                tx_id=tx_id, tx_model=tx_model,
                genome=genome, anno=anno,
                cleavage_params=cleavage_params,
                w2f_reassignment=args.w2f_reassignment,
                sec_truncation=args.selenocysteine_termination
            )
        except:
            logger.error('Exception raised from %s', tx_id)
            raise

        for peptide in peptides:
            peptide_pool.add_peptide(
                peptide=peptide,
                canonical_peptides=canonical_peptides,
                cleavage_params=cleavage_params
            )

    peptide_pool.write(args.output_path)

    logger.info('Alternative translation peptide FASTA file written to disk.')

def call_alt_translation_main(tx_id:str, tx_model:TranscriptAnnotationModel,
        genome:DNASeqDict, anno:GenomicAnnotation,
        cleavage_params:params.CleavageParams,
        w2f_reassignment:bool, sec_truncation:bool):
    """ wrapper of graph operations to call peptides """
    chrom = tx_model.transcript.chrom
    tx_seq = tx_model.get_transcript_sequence(genome[chrom])

    dgraph = svgraph.ThreeFrameTVG(
        seq=tx_seq,
        _id=tx_id,
        cds_start_nf=tx_model.is_cds_start_nf(),
        has_known_orf=True,
        mrna_end_nf=tx_model.is_mrna_end_nf(),
        cleavage_params=cleavage_params,
        coordinate_feature_type='transcript',
        coordinate_feature_id=tx_id
    )
    dgraph.gather_sect_variants(anno)
    dgraph.init_three_frames()
    pgraph = dgraph.translate()
    pgraph.create_cleavage_graph()
    peptide_anno = pgraph.call_variant_peptides(
        check_variants=True,
        truncate_sec=sec_truncation,
        w2f=w2f_reassignment,
        check_external_variants=False
    )
    peptide_map:Dict[Seq, Set[str]] = {}
    for seq, annotated_labels in peptide_anno.items():
        for label in annotated_labels:
            if seq in peptide_map:
                peptide_map[seq].add(label.label)
            else:
                peptide_map[seq] = {label.label}
    peptides = set()
    for seq, labels in peptide_map.items():
        label = VARIANT_PEPTIDE_SOURCE_DELIMITER.join(labels)
        peptides.add(
            aa.AminoAcidSeqRecord(
                seq=seq,
                description=label,
                name=label
            )
        )
    return peptides



""" Model for DNA sequence data
"""
from Bio import SeqIO
from moPepGen.dna import DNASeqRecord


class DNASeqDict(dict):
    """ A DNASeqDict object is a dict-like object that the values are
    DNASeqRecord objects.
    """
    def __init__(self, *args, **kwargs) -> None:
        for val in kwargs.values():
            self._validate(val)
        super().__init__(*args, **kwargs)

    @staticmethod
    def _validate(val):
        """ validate values """
        if not isinstance(val, DNASeqRecord):
            raise TypeError(
                "'DNASeqDict' only accepts 'DNASeqRecord' objects."
            )

    def __setitem__(self, k:str, v:DNASeqRecord) -> None:
        """ Set items. Only allow DNASeqRecord in values. """
        self._validate(v)
        super().__setitem__(k, v)

    def dump_fasta(self, path:str) -> None:
        """ Dump a FASTA file to a DNASeqDict

        Args:
            path (str): Path to the genome assembly FASTA file.
        """
        for record in SeqIO.parse(path, 'fasta'):
            record.__class__ = DNASeqRecord
            record.id = record.id.split(' ')[0]
            record.id = record.id.split('|')[0]
            if record.id in self.keys():
                raise ValueError(
                    'Duplicated seqnames found in FASTA file: ' + path
                )
            self[record.id] = record

    def with_chr(self) -> bool:
        """ Check whether keys have 'chr' or not """
        # so far only works with humans
        return bool('chr1' in self.keys())



""" `splitFasta` takes the FASTA file with variant peptide sequences called by
[`callVariant`](./call-variant.md) with or without novel ORF peptides called by
[`callNovelORF`](./call-novel-orf.md), and splits peptide sequences into databases.
The split database FASTA files can be used for sequential library searching. """
from __future__ import annotations
import argparse
from pathlib import Path
from moPepGen.aa import PeptidePoolSplitter
from moPepGen import SPLIT_DATABASE_KEY_SEPARATER, get_logger
from moPepGen.cli import common


GVF_FILE_FORMAT = ['.gvf']
FASTA_FILE_FORMAT = ['.fasta', '.fa']


# pylint: disable=W0212
def add_subparser_split_fasta(subparser:argparse._SubParsersAction):
    """ CLI for moPepGen splitFasta """
    p:argparse.ArgumentParser = subparser.add_parser(
        name='splitFasta',
        help='Split variant peptide FASTA database generated by moPepGen.',
        description='Split variant peptide FASTA database generated by'
        ' moPepGen into separate files.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    p.add_argument(
        '--gvf',
        type=Path,
        help='File path to GVF files. All GVF files must be generated by'
        f" moPepGen parsers. Valid formats: {GVF_FILE_FORMAT}",
        metavar='<files>',
        nargs='+'
    )
    p.add_argument(
        '--variant-peptides',
        type=Path,
        help='File path to the variant peptide FASTA database file. Must be'
        f" generated by moPepGen callVariant. Valid formats: {FASTA_FILE_FORMAT}",
        metavar='<file>',
        default=None
    )
    p.add_argument(
        '--novel-orf-peptides',
        type=Path,
        help='File path to the novel ORF peptide FASTA database file. Must be'
        f" generated by moPepGen callNovelORF. Valid formats: {FASTA_FILE_FORMAT}",
        metavar='<file>',
        default=None
    )
    p.add_argument(
        '--alt-translation-peptides',
        type=Path,
        help='File path to the alt translation peptide FASTA file. Must be'
        f"generated by moPepGen callAltTranslation. Valid formats: {FASTA_FILE_FORMAT}",
        metavar='<file>',
        default=None
    )
    p.add_argument(
        '-o', '--output-prefix',
        type=Path,
        help='Output prefix',
        metavar = '<value>'
    )
    p.add_argument(
        '--order-source',
        type=str,
        help='Order of sources, separate by comma. E.g., SNP,SNV,Fusion',
        metavar='<value>'
    )
    p.add_argument(
        '--group-source',
        type=str,
        help='Group sources. The peptides with sources grouped will be written'
        ' to the same FASTA file. E.g., "PointMutation:gSNP,sSNV'
        ' INDEL:gINDEL,sINDEL".',
        metavar='<value>',
        nargs='*'
    )
    p.add_argument(
        '--max-source-groups',
        type=int,
        help='Maximal number of different source groups to be separate into'
        ' individual database FASTA files. Defaults to 1',
        default=1,
        metavar='<number>'
    )
    p.add_argument(
        '--additional-split',
        type=str,
        help='For peptides that were not already split into FASTAs up to'
        ' max_source_groups, those involving the following source will be split'
        ' into additional FASTAs with decreasing priority. E.g., '
        " 'gSNP-NovelORF', 'gSNP-NovelORF gSNP-gINDEL'",
        metavar='<value>',
        default=None,
        nargs="*"
    )

    common.add_args_reference(p, genome=False, proteome=True)
    common.add_args_debug_level(p)
    common.print_help_if_missing_args(p)
    p.set_defaults(func=split_fasta)
    return p

def split_fasta(args:argparse.Namespace) -> None:
    """ Split peptide database """
    logger = get_logger()
    for file in args.gvf:
        common.validate_file_format(
            file, GVF_FILE_FORMAT, check_readable=True
        )
    peptide_fastas = [args.variant_peptides, args.novel_orf_peptides, args.alt_translation_peptides]
    if all(x is None for x in peptide_fastas):
        raise ValueError(
            "At least one of '--variant-peptide', '--novel-orf-peptide' or"
            "' --alt-translation-peptide' must be given. Please verify your command."
        )
    for file in peptide_fastas:
        if file is not None:
            common.validate_file_format(
                file, FASTA_FILE_FORMAT, check_readable=True
            )

    common.print_start_message(args)

    _, anno, *_ = common.load_references(args, load_genome=False, \
        load_proteome=True, load_canonical_peptides=False,
        check_protein_coding=True)

    tx2gene = {}
    coding_tx = set()
    for tx_id in anno.transcripts:
        tx_model = anno.transcripts[tx_id]
        tx2gene[tx_id] = tx_model.transcript.gene_id
        if tx_model.is_protein_coding:
            coding_tx.add(tx_id)
    del anno

    if args.order_source:
        source_order = {}
        for i,val in enumerate(args.order_source.split(',')):
            if '-' in val:
                source = frozenset(val.split('-'))
            else:
                source = val
            if source in source_order:
                raise ValueError(
                    f"Non-unique value found from `--group-source`: {source}"
                )
            source_order[source] = i
    else:
        source_order = None

    group_map = None
    if args.group_source:
        group_map = {}
        for it in args.group_source:
            key, val = it.split(':')
            for v in val.split(','):
                group_map[v] = key

    splitter = PeptidePoolSplitter(order=source_order, group_map=group_map)

    if args.variant_peptides:
        with open(args.variant_peptides, 'rt') as handle:
            splitter.load_database(handle)
        logger.info("Variant FASTA loaded: %s.", args.variant_peptides)

    if args.novel_orf_peptides:
        with open(args.novel_orf_peptides, 'rt') as handle:
            splitter.load_database(handle)
        logger.info("Novel ORF FASTA loaded: %s", args.novel_orf_peptides)

    if args.alt_translation_peptides:
        with open(args.alt_translation_peptides, 'rt') as handle:
            splitter.load_database(handle)
        logger.info("Alternative Translation FASTA loaded: %s", args.alt_translation_peptides)

    for file in args.gvf:
        with open(file, 'rt') as handle:
            splitter.load_gvf(handle)
        logger.info("GVF file used: %s", file)

    additional_split = args.additional_split or []
    sep = SPLIT_DATABASE_KEY_SEPARATER
    additional_split = [set(x.split(sep)) for x in additional_split]

    logger.info("Using source order: %s", splitter.order)
    logger.info("Using source group: %s", splitter.get_reversed_group_map())

    logger.info("Start splitting...")

    splitter.split(
        max_groups=args.max_source_groups,
        additional_split=additional_split,
        tx2gene=tx2gene,
        coding_tx=coding_tx
    )

    logger.info('Database split finished')

    splitter.write(args.output_prefix)

    logger.info('Split databases saved to disk.')



""" Variant Record Pool """
from __future__ import annotations
import copy
from typing import Dict, IO, Iterable, List, TYPE_CHECKING, Union
from moPepGen import ERROR_INDEX_IN_INTRON, circ
from . import VariantRecord, io, GVFMetadata
from .VariantRecordPoolOnDisk import TranscriptionalVariantSeries


# To avoid circular import
if TYPE_CHECKING:
    from moPepGen.gtf import GenomicAnnotation
    from moPepGen.dna import DNASeqDict

T = Dict[str, List[VariantRecord]]

class VariantRecordPool():
    """ This is used to store all variant records read in from GVF files.
    Variant records (VariantRecord) are first filtered based on their location
    and put into different buckets of the pool.

    Attributes:
        transcriptional (Dict[str, List[VariantRecord]]): Variant records
            located in the transcript regions (exons), in transcriptional
            coordinates.
        intronic (Dict[str, List[VariantRecord]]): Variant records located in
            the intronic regions. In gene coordinates.
        genetic (Dict[str, List[VariantRecord]]): Variant records without a
            transcript ID (e.g., UTR). In gene coordinates.
    """
    def __init__(self, data:Dict[str, TranscriptionalVariantSeries]=None,
            anno:GenomicAnnotation=None):
        """ Constructor

        Args:
            transcriptional (Dict[str, List[VariantRecord]]): Variant records
                located in the transcript regions (exons), in transcriptional
                coordinates.
            intronic (Dict[str, List[VariantRecord]]): Variant records located
                in the intronic regions. In gene coordinates.
            genetic (Dict[str, List[VariantRecord]]): Variant records without a
                transcript ID (e.g., UTR). In gene coordinates.
        """
        self.data = data or {}
        self.anno = anno

    def __contins__(self, key:str):
        """ in """
        return key in self.data

    def __getitem__(self, key:str) -> Union[TranscriptionalVariantSeries]:
        """ get item """
        return self.data[key]

    def __setitem__(self, key:str, val:Union[TranscriptionalVariantSeries]):
        """ setitem """
        self.data[key] = val

    def __iter__(self) -> Iterable[str]:
        """ generator """
        for key in self.data:
            yield key

    def __copy__(self) -> VariantRecordPool:
        """ copy """
        data = copy.copy(self.data)
        anno = copy.copy(self.anno)
        return self.__class__(data, anno)

    def add_intronic_variant(self, record:VariantRecord, tx_id:str=None):
        """ Add a variant with genetic coordinate that is in the intron of
        a transcript """
        if tx_id is None:
            tx_id = record.transcript_id
        if tx_id not in self:
            self[tx_id] = TranscriptionalVariantSeries()
        self[tx_id].intronic.append(record)

    def add_intronic_variants(self, records:Iterable[VariantRecord], tx_id:str):
        """ Add multiple intronic VariantRecord with genetic coordinate """
        for record in records:
            self.add_intronic_variant(record=record, tx_id=tx_id)

    def add_transcriptional_variant(self, record:VariantRecord, tx_id:str=None):
        """ Add a variant with transcriptional coordinate """
        if tx_id is None:
            tx_id = record.location.seqname
        if tx_id not in self:
            self[tx_id] = TranscriptionalVariantSeries()
        self[tx_id].transcriptional.append(record)

    def add_fusion_variant(self, record:VariantRecord, tx_id:str=None):
        """ Add a fuction variant """
        if tx_id is None:
            tx_id = record.location.seqname
        if tx_id not in self:
            self[tx_id] = TranscriptionalVariantSeries()
        self[tx_id].fusion.append(record)

    def add_circ_rna(self, record:circ.CircRNAModel, tx_id:str=None):
        """ Add a circRNA """
        if tx_id is None:
            tx_id = record.transcript_id
        if tx_id not in self:
            self[tx_id] = TranscriptionalVariantSeries()
        self[tx_id].circ_rna.append(record)

    def load_variants(self, handle:IO, anno:GenomicAnnotation,
            genome:DNASeqDict):
        """ Load variants """
        handle.seek(0)
        metadata = GVFMetadata.parse(handle)
        handle.seek(0)

        if metadata.is_circ_rna():
            for record in circ.io.parse(handle):
                self.add_circ_rna(record)
            return

        for record in io.parse(handle):
            tx_id = record.transcript_id
            if record.is_fusion():
                record.shift_breakpoint_to_closest_exon(anno)
                tx_record = record.to_transcript_variant(anno, genome, tx_id)
                self.add_fusion_variant(tx_record, tx_id)
                continue

            if record.is_spanning_over_splicing_site(anno, tx_id):
                continue

            try:
                tx_record = record.to_transcript_variant(anno, genome, tx_id)
                if tx_record.type == 'Deletion':
                    tx_model = anno.transcripts[tx_id]
                    chrom = tx_model.transcript.chrom
                    tx_seq = tx_model.get_transcript_sequence(genome[chrom])
                    tx_record.shift_deletion_up(tx_seq)
                self.add_transcriptional_variant(tx_record, tx_id)

            except ValueError as e:
                if e.args[0] == ERROR_INDEX_IN_INTRON:
                    self.add_intronic_variant(record, tx_id)
                else:
                    raise e
        self.sort()

    def sort(self):
        """ sort """
        for key in self:
            series = self[key]
            if isinstance(series, TranscriptionalVariantSeries):
                series.sort()

    def filter_variants(self, gene_id:str=None, tx_ids:List[str]=None,
            exclude_type:List[str]=None, start:int=None, end:int=None,
            intron:bool=True, segments:Iterable[VariantRecord]=None,
            return_coord:str='gene') -> List[VariantRecord]:
        """ Filter variants of located at a given position (start and end) of
        a gene.

        Arguments:
            gene_id (str): Gene ID
            start (int): Start position with genetic coordinates.
            end (int): End position with genetic coordinates. If
            exclude_type (List[str]): Variant types that should be excluded.
        """
        if return_coord not in ['gene', 'transcript']:
            raise ValueError(
                "Don't know how to return variants in coordinate of "
                f"{return_coord}. return_coord must be either 'gene' or "
                "'transcript'"
            )
        if return_coord == 'transcript' and intron is True:
            raise ValueError(
                "Don't know how to return intronic variants in transcript"
                "coordinates."
            )
        if segments:
            def _filter(x):
                for segment in segments:
                    if segment.location.start < x.location.start < \
                            x.location.end < segment.location.end:
                        return True
                return False
        elif start is not None and end is not None:
            _filter = lambda x: x.location.start > start and x.location.end < end
        elif start is not None:
            _filter = lambda x: x.location.start > start
        elif end is not None:
            _filter = lambda x: x.location.end < end
        else:
            raise ValueError('Arguments provided do not match requirement.')

        records = set()
        if tx_ids:
            _tx_ids = set(tx_ids)
        else:
            _tx_ids = set()
        if gene_id:
            _tx_ids.update(self.anno.genes[gene_id].transcripts)
        for tx_id in _tx_ids:
            gene_id = self.anno.transcripts[tx_id].transcript.gene_id
            if tx_id in self:
                for record in self[tx_id].transcriptional:
                    if record.type in exclude_type:
                        continue
                    try:
                        record_gene = self.anno.variant_coordinates_to_gene(record, gene_id)
                    except ValueError as e:
                        if record.is_merged_mnv():
                            continue
                        raise e
                    if _filter(record_gene):
                        if return_coord == 'gene':
                            records.add(record_gene)
                        else:
                            records.add(record)
            if not intron:
                continue
            if tx_id in self:
                for record in self[tx_id].intronic:
                    if record.type in exclude_type:
                        continue
                    if _filter(record):
                        if return_coord == 'gene':
                            records.add(record)
        records = list(records)
        records.sort()
        return records



""" moPepGen util module """
from moPepGen.util import brute_force
from moPepGen.util import downsample_reference
from moPepGen.util import validate_variant_calling
from moPepGen.util import brute_force_novel_orf
from moPepGen.util import fuzz_test
from moPepGen.util import extract_gvf
from moPepGen.util import validate_novel_orf_calling
from moPepGen.util import add_fuzz_test_log

PROG_NAME = 'moPepGen-util'



""" module for peptide pool summarizer """
from __future__ import annotations
import itertools
import statistics
from typing import Dict, IO, List, Set, FrozenSet, Tuple, Optional
import matplotlib.pyplot as plt
from moPepGen import seqvar, constant
from moPepGen.aa.AminoAcidSeqRecord import AminoAcidSeqRecord
from moPepGen.aa.VariantPeptideLabel import VariantPeptideInfo, \
    VariantSourceSet, LabelSourceMapping
from moPepGen.aa.VariantPeptidePool import VariantPeptidePool
from moPepGen.seqvar.GVFMetadata import GVFMetadata


SOURCES_INTERNAL = [
    constant.SOURCE_NOVEL_ORF,
    constant.SOURCE_SEC_TERMINATION,
    constant.SOURCE_CODON_REASSIGNMENT
]

MUTUALLY_EXCLUSIVE_PARSERS:Dict[str,List[str]] = {
    'parseSTARFusion': [
        'parseFusionCatcher',
        'parseArriba',
        'parseCIRCexplorer',
        'parseRMATS'
    ],
    'parseFusionCatcher': [
        'parseSTARFusion',
        'parseArriba',
        'parseCIRCexplorer',
        'parseRMATS'
    ],
    'parseArriba': [
        'parseSTARFusion',
        'parseFusionCatcher',
        'parseCIRCexplorer',
        'parseRMATS'
    ],
    'parseCIRCexplorer': [
        'parseSTARFusion',
        'parseFusionCatcher',
        'parseArriba',
        'parseRMATS'
    ],
    'parseRMATS': [
        'parseSTARFusion',
        'parseFusionCatcher',
        'parseArriba',
        'parseCIRCexplorer'
    ]
}

# BPF default color scheme.
# Hopefully no one is using more than 11 miscleavages
COLOR_SCHEME = [
    "#FFA500", "#458B00", "#68228B", "#FFD700", "#1E90FF", "#CD2626", "#9ACD32"
    "#FF7F00", "#473C8B", "#43CD80", "#CD3278", "#00C5CD"
]

class NoncanonicalPeptideSummaryTable():
    """ Summary table for noncaonincal peptides

    Attributes:
        data (Dict[FrozonSet[str], Dict[str,int]]): Mapping from variant source
            to number of miscleavages to count of peptides.
        max_misc (int): Max number of miscleavages allowed.
    """
    KEY_N_TOTAL = 'n_total'
    def __init__(self, data:Dict[FrozenSet[str], Dict[str,int]]=None,
            sources:Set[str]=None, max_misc:int=0):
        """ constructor """
        # mapping: source -> misc -> count
        self.data = data or {}
        self.sources = sources or set()
        self.max_misc = max_misc

    @staticmethod
    def get_key_x_misc(x:int):
        """ get the table column name for number of peptides per combination of
        sources with x number of miscleavages """
        return f"n_{x}_misc"

    def get_n_total(self, key:str) -> int:
        """ gety the table column name for total number of peptides per
        combination of sources """
        return self.data[key][self.KEY_N_TOTAL]

    def get_n_x_misc(self, key:str, misc:int) -> int:
        """ get the number of peptides with x number of miscalvages from the
        summary table """
        entry = self.data[key]
        key = self.get_key_x_misc(misc)
        return entry.get(key, 0)

    def add_entry(self,
            seq:AminoAcidSeqRecord,
            label_map:LabelSourceMapping,
            group_map:Dict[str,str],
            tx2gene:Dict[str,str],
            coding_tx:Set[str],
            enzyme:str) -> None:
        """ Add peptide entry to the summary table """
        peptide_labels = VariantPeptideInfo.from_variant_peptide(
            peptide=seq, tx2gene=tx2gene, coding_tx=coding_tx,
            label_map=label_map, group_map=group_map,
            check_source=True
        )
        peptide_labels.sort()
        sources = frozenset(peptide_labels[0].sources)
        self.increment_total(sources)

        exception = 'trypsin_exception' if enzyme == 'trypsin' else None
        misc = len(seq.find_all_enzymatic_cleave_sites(enzyme, exception))
        self.increment_misc(sources, misc)

    def add_source(self, source:str) -> None:
        """ add source """
        self.sources.add(source)

    def _increment(self, sources:FrozenSet[str], key):
        """ Increment """
        for source in sources:
            self.add_source(source)

        if sources not in self.data:
            self.data[sources] = {key: 1}
        elif key not in self.data[sources]:
            self.data[sources][key] = 1
        else:
            self.data[sources][key] += 1

    def increment_total(self, sources:FrozenSet[str]):
        """ Increment the total number of peptides for the given combination of
        sources """
        self._increment(sources, self.KEY_N_TOTAL)

    def increment_misc(self, sources:FrozenSet[str], misc:int):
        """ Increment the number of peptides with x miscleavages for the given
        combination of sources """
        self.max_misc = max(self.max_misc, misc)
        self._increment(sources, self.get_key_x_misc(misc))

    def get_keys(self) -> Tuple[str]:
        """ Get all column names """
        return ('n_total', *[self.get_key_x_misc(i) for i in range(self.max_misc + 1)])

    def get_stringified_summary_entry(self, key:str, order:Dict[str,int],
            sep:str='\t') -> str:
        """ Get a row from the summary table for a given combination of sources. """
        rowname = '-'.join(sorted(key, key=lambda x: order[x]))
        if key not in self.data:
            entry = [rowname, '0']
            for _ in range(self.max_misc + 1):
                entry.append('0')
        else:
            entry = [rowname, str(self.get_n_total(key))]
            for x in range(self.max_misc + 1):
                entry.append(str(self.get_n_x_misc(key, x)))
        return sep.join(entry)

    def has_source(self, source:str) -> bool:
        """ Checks if the given source exist in the summary table """
        return source in self.sources

class PeptidePoolSummarizer():
    """ Summarize the variant peptide pool called by moPepGen callVariant. """
    def __init__(self,
            peptides:VariantPeptidePool=None,
            summary_table:NoncanonicalPeptideSummaryTable=None,
            label_map:LabelSourceMapping=None,
            order:Dict[str,int]=None,
            group_map:Dict[str,str]=None,
            source_parser_map:Dict[str,str]=None,
            ignore_missing_source:bool=False):
        """ """
        self.peptides = peptides
        self.summary_table = summary_table or NoncanonicalPeptideSummaryTable()
        self.label_map = label_map or LabelSourceMapping()
        self.order = order or {}
        self.group_map = group_map or {}
        self.source_parser_map = source_parser_map or {}
        self.ignore_missing_source = ignore_missing_source

    def get_reversed_group_map(self) -> Dict[str, List[str]]:
        """ Reverse group map """
        group_map:Dict[str, List[str]] = {}
        for k,v in self.group_map.items():
            if v in group_map:
                group_map[v].append(k)
            else:
                group_map[v] = [k]
        return group_map

    def get_parsers_from_source(self, source:str) -> Set[str]:
        """ Get parsers from source """
        group_map = self.get_reversed_group_map()
        parsers = set()
        # pylint: disable=R1715
        if source in group_map:
            sources = group_map[source]
        else:
            sources = [source]

        for s in sources:
            if s in SOURCES_INTERNAL:
                continue
            parser = self.source_parser_map[s]
            parsers.add(parser)
        return parsers

    def append_order_internal_sources(self):
        """ Add internal sources that are not present in any GTFs, including
        novel ORF, sec termination, and codon reassignment. """
        for source in SOURCES_INTERNAL:
            if source in self.group_map:
                source = self.group_map[source]
            if source not in self.order:
                self.append_order(source)

    def append_order(self, source:str):
        """ Add a group to the end of the order. """
        if source in self.order:
            return
        if source in self.group_map:
            source = self.group_map[source]
        if source in self.order:
            return
        self.order[source] = max(self.order.values()) + 1 if self.order else 0

    def load_database(self, handle:IO) -> None:
        """ load peptide database """
        pool = VariantPeptidePool.load(handle)
        if not self.peptides:
            self.peptides = pool
            return
        for peptide in pool.peptides:
            self.peptides.add_peptide(peptide, None, skip_checking=True)

    def update_label_map(self, handle:IO):
        """ Read GVF files and add to the label source map. """
        metadata = GVFMetadata.parse(handle)
        self.append_order(metadata.source)
        self.source_parser_map[metadata.source] = metadata.parser
        for gene_id, _, label in seqvar.io.parse_label(handle):
            self.label_map.add_record(gene_id, label, metadata.source)

    def count_peptide_source(self, tx2gene:Dict[str,str], coding_tx:Set[str], enzyme:str):
        """ Count number of peptides in each source or combinations of sources. """
        VariantSourceSet.set_levels(self.order)
        for seq in self.peptides.peptides:
            self.summary_table.add_entry(
                seq=seq,
                label_map=self.label_map,
                group_map=self.group_map,
                tx2gene=tx2gene,
                coding_tx=coding_tx,
                enzyme=enzyme
            )

    def contains_exclusive_sources(self, sources:Set[str]) -> bool:
        """ Checks whether the given sources contains any mutually exclusive
        parsers. """
        group_map = self.get_reversed_group_map()
        ungrouped_sources = set()
        for source in sources:
            if source in group_map:
                ungrouped_sources.update(group_map[source])
            else:
                ungrouped_sources.add(source)

        # Two source groups are exclusive only when every source in a group A
        # has at least one mutually exclusive group in group B, and also the
        # other way around.

        for source in sources:
            parsers = self.get_parsers_from_source(source)
            if not parsers:
                continue

            for other in sources:
                other_parsers = self.get_parsers_from_source(other)
                if not other_parsers:
                    continue
                is_exclusive = True
                for x in parsers:
                    if x not in MUTUALLY_EXCLUSIVE_PARSERS:
                        is_exclusive = False
                        break
                    for y in other_parsers:
                        if y not in MUTUALLY_EXCLUSIVE_PARSERS:
                            is_exclusive = False
                            break
                        if x not in MUTUALLY_EXCLUSIVE_PARSERS[y]:
                            is_exclusive = False
                            break

                if is_exclusive:
                    return True

        return False

    def write_summary_table(self, handle:IO):
        """ Write summary table to output. """
        summary_keys = self.summary_table.get_keys()
        header = '\t'.join(['sources', *summary_keys])
        handle.write(header + '\n')
        sources = [it[0] for it in sorted(self.order.items(), key=lambda x:x[1])
                    if not isinstance(it[0], frozenset)]
        for i in range(len(sources)):
            for comb in itertools.combinations(sources, i + 1):
                if self.ignore_missing_source:
                    # Ignore it if the source isn't present in any GVF given.
                    if any(not self.summary_table.has_source(k) for k in comb):
                        continue
                if self.contains_exclusive_sources(comb):
                    continue
                key = frozenset(comb)
                record = self.summary_table.get_stringified_summary_entry(key, self.order)
                handle.write(record + '\n')

    def create_barplot(self, width:float=6, height:float=8, ax:Optional[plt.Axes]=None,
            scale:str=None) -> plt.Axes:
        """ Make a barplot of the summarized data. """
        # misc -> source -> count
        data:Dict[int,List[int]] = {}
        keys:List[str] = []
        sources = [it[0] for it in sorted(self.order.items(), key=lambda x:x[1])]
        for i in range(len(sources)):
            for comb in itertools.combinations(sources, i + 1):
                if self.ignore_missing_source:
                    # Ignore it if the source isn't present in any GVF given.
                    if any(not self.summary_table.has_source(k) for k in comb):
                        continue
                if self.contains_exclusive_sources(comb):
                    continue
                key = frozenset(comb)
                if key not in self.summary_table.data:
                    continue

                keys.append('-'.join(key))
                for x in range(self.summary_table.max_misc + 1):
                    n = self.summary_table.get_n_x_misc(key, x)
                    if x not in data:
                        data[x] = []
                    data[x].append(n)

        totals = []
        for x in data.values():
            if not totals:
                totals = x
            else:
                totals = [i+j for i,j in zip(totals, x)]
                mean = statistics.mean(totals)
                median = statistics.median(totals)

        # Unless specified, log scale is used if the summary distribution is skewed.
        if scale not in ['normal', 'log']:
            scale = 'normal' if mean / median <= 2.5 else 'log'

        if ax is None:
            _, ax = plt.subplots(figsize=(width, height))

        if scale == 'log':
            ax.barh(
                y=list(reversed(keys)), width=list(reversed(totals))
            )
            ax.set_xscale('log')
        else:
            offset = None
            for i,n in enumerate(data.keys()):
                vals = list(reversed(data[n]))
                if offset:
                    vals = [x + y for x,y in zip(vals, offset)]
                ax.barh(
                    y=list(reversed(keys)), width=vals, left=offset, label=n,
                    color=COLOR_SCHEME[i]
                )
                offset = vals
                ax.legend(title='Miscleavages')
        ax.set_xlabel('Number of Peptides')
        return ax



""" `generateIndex` takes the reference genome FASTA, annotation GTF, and the
translated proteome FASTA file, and processes them so they can be read by
subsequent moPepGen commands quickly. The outputted index files also contain the
canonical peptide pool. The index files can then be used in any moPepGen
command. It is recommended to run `generateIndex` before any analysis using
moPepGen to avoid processing the reference files repeatedly and save massive
time. """
from __future__ import annotations
import argparse
from pathlib import Path
import sys
from moPepGen import dna, aa, params, get_logger
from moPepGen.index import IndexDir
from moPepGen.cli import common


# pylint: disable=W0212
def add_subparser_generate_index(subparsers:argparse._SubParsersAction):
    """ CLI for moPepGen generateIndex """
    p:argparse.ArgumentParser = subparsers.add_parser(
        name='generateIndex',
        help='Generate genome and proteome index files for moPepGen',
        description='Generate genome and proteome index files for moPepGen'
        ' parsers and peptide caller.',
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    p.add_argument(
        '-o', '--output-dir',
        type=Path,
        help='Ouput directory for index files.',
        metavar='<file>',
        dest='output_dir',
        required=True
    )
    p.add_argument(
        '--gtf-symlink',
        help='Create a symlink of the GTF file instead of copying it.',
        action='store_true'
    )
    p.add_argument(
        '-f', '--force',
        action='store_true',
        help='Force write data to index dir.'
    )
    common.add_args_reference(p, index=False)
    common.add_args_cleavage(p)
    common.add_args_debug_level(p)
    p.set_defaults(func=generate_index)
    common.print_help_if_missing_args(p)
    return p

def generate_index(args:argparse.Namespace):
    """ Generate index  """
    path_genome:Path = args.genome_fasta
    path_gtf:Path = args.annotation_gtf
    parth_proteome:Path = args.proteome_fasta

    rule:str = args.cleavage_rule
    miscleavage:int = int(args.miscleavage)
    min_mw:float = float(args.min_mw)
    min_length:int = int(args.min_length)
    max_length:int = int(args.max_length)
    exception = args.cleavage_exception
    invalid_protein_as_noncoding:bool = args.invalid_protein_as_noncoding

    output_dir:Path = args.output_dir

    common.print_start_message(args)
    logger = get_logger()

    output_dir.mkdir(exist_ok=True)
    index_dir = IndexDir(output_dir)
    if any(index_dir.path.iterdir()):
        if args.force:
            index_dir.wipe_canonical_peptides()
            index_dir.init_metadata()
        else:
            logger.error("Index directory already exists.")
            sys.exit(1)

    # genome fasta
    genome = dna.DNASeqDict()
    genome.dump_fasta(path_genome)
    logger.info('Genome FASTA loaded')
    index_dir.save_genome(genome)
    logger.info('Genome FASTA saved to disk.')

    # proteome
    proteome = aa.AminoAcidSeqDict()
    proteome.dump_fasta(parth_proteome, source=args.reference_source)
    logger.info('Proteome FASTA loaded.')
    index_dir.save_proteome(proteome)
    logger.info('Proteome FASTA saved to disk.')

    # annotation GTF
    anno = index_dir.save_annotation(
        file=path_gtf,
        source=args.reference_source,
        proteome=proteome,
        invalid_protein_as_noncoding=invalid_protein_as_noncoding,
        symlink=args.gtf_symlink
    )
    logger.info('Genome annotation GTF saved to disk.')

    # canoincal peptide pool
    canonical_peptides = proteome.create_unique_peptide_pool(
        anno=anno, rule=rule, exception=exception, miscleavage=miscleavage,
        min_mw=min_mw, min_length = min_length, max_length = max_length
    )
    cleavage_params = params.CleavageParams(
        enzyme=rule, exception=exception, miscleavage=miscleavage,
        min_mw=min_mw, min_length = min_length, max_length = max_length
    )
    logger.info('canonical peptide pool generated.')
    index_dir.save_canonical_peptides(canonical_peptides, cleavage_params)
    logger.info('canonical peptide pool saved to disk.')

    # create list of coding transcripts
    coding_tx = {tx_id for tx_id, tx_model in anno.transcripts.items()
        if tx_model.is_protein_coding}
    index_dir.save_coding_tx(coding_tx)

    # save metadata
    index_dir.metadata.source = anno.source
    index_dir.save_metadata()
    logger.info('metadata saved.')



""" Module for Arriba parser """
from __future__ import annotations
import itertools
from typing import Iterable, List, IO
from moPepGen.SeqFeature import FeatureLocation
from moPepGen import seqvar, gtf, dna, err


def parse(handle:IO) -> Iterable[ArribaRecord]:
    """ parse Arriba's output """
    for line in handle:
        line:str
        if line.startswith('#'):
            continue
        fields = line.rstrip().split('\t')
        fields[9] = int(fields[9])
        fields[10] = int(fields[10])
        fields[11] = int(fields[11])
        fields[12] = int(fields[12])
        fields[13] = int(fields[13])
        fields[14] = ArribaConfidence(fields[14])
        fields[26] = fields[26].split(',')
        fields[29] = fields[29].split(',')

        yield ArribaRecord(*fields)

class ArribaConfidence():
    """ Arriba's confidence """
    levels = {'low': 0, 'medium': 1, 'high': 2}
    """ Arriba's confidence """
    def __init__(self, data:str):
        """ constructor """
        if data not in self.levels:
            raise ValueError(f"Invalid confidence of {data}")
        self.data = data

    def to_int(self) -> int:
        """ Convert to int """
        return self.levels[self.data]

    def __lt__(self, other:ArribaConfidence) -> bool:
        """ larger than """
        return self.to_int() > other.to_int()

    def __le__(self, other:ArribaConfidence) -> bool:
        """ larger or equal to """
        return self == other or self > other

    def __st__(self, other:ArribaConfidence) -> bool:
        """ smaller than """
        return not self >= other

    def __se__(self, other:ArribaConfidence) -> bool:
        """ smaller or equal to """
        return not self > other

    def __eq__(self, other:ArribaConfidence) -> bool:
        """ equal to """
        return self.data == other.data

    def __ne__(self, other:ArribaConfidence) -> bool:
        """ not equal to """
        return not self == other

class ArribaRecord():
    """ Arriba Record. More info see:
    https://arriba.readthedocs.io/en/latest/output-files/ """
    def __init__(self, gene1:str, gene2:str, strand1:str, strand2:str,
            breakpoint1:str, breakpoint2:str, site1:str, site2:str, _type:str,
            split_reads1:str, split_reads2:str, discordant_mates:int,
            coverage1:int, coverage2:int, confidence:ArribaConfidence,
            reading_frame:str, tags:str, retained_protein_domains:str,
            closest_genomic_breakpoint1:str, closest_genomic_breakpoint2:str,
            gene_id1:str, gene_id2:str, transcript_id1:str, transcript_id2:str,
            direction1:str, direction2:str, filters:List[str],
            fusion_transcript:str, peptide_sequence:str,
            read_identifiers:List[str]):
        """ Constructor """
        self.gene1 = gene1
        self.gene2 = gene2
        self.strand1 = strand1
        self.strand2 = strand2
        self.breakpoint1 = breakpoint1
        self.breakpoint2 = breakpoint2
        self.site1 = site1
        self.site2 = site2
        self._type = _type
        self.split_reads1 = split_reads1
        self.split_reads2 = split_reads2
        self.discordant_mates = discordant_mates
        self.coverage1 = coverage1
        self.coverage2 = coverage2
        self.confidence = confidence
        self.reading_frame = reading_frame
        self.tags = tags
        self.retained_protein_domains = retained_protein_domains
        self.closest_genomic_breakpoint1 = closest_genomic_breakpoint1
        self.closest_genomic_breakpoint2 = closest_genomic_breakpoint2
        self.gene_id1 = gene_id1
        self.gene_id2 = gene_id2
        self.transcript_id1 = transcript_id1
        self.transcript_id2 = transcript_id2
        self.direction1 = direction1
        self.direction2 = direction2
        self.filters = filters
        self.fusion_transcript = fusion_transcript
        self.peptide_sequence = peptide_sequence
        self.read_identifiers = read_identifiers

    @property
    def gene_strand1(self) -> int:
        """ gene strand1 """
        return self.infer_strand(self.strand1.split('/')[0])

    @property
    def gene_strand2(self) -> int:
        """ gene strand2 """
        return self.infer_strand(self.strand2.split('/')[0])

    @property
    def transcript_strand1(self) -> int:
        """ transcript strand1 """
        return self.infer_strand(self.strand1.split('/')[1])

    @property
    def transcript_strand2(self) -> int:
        """ transcript strand2 """
        return self.infer_strand(self.strand2.split('/')[1])

    @property
    def breakpoint1_position(self) -> int:
        """ Get breakpoint1 position """
        return int(self.breakpoint1.split(':')[1])

    @property
    def breakpoint2_position(self) -> int:
        """ Get breakpoint2 position """
        return int(self.breakpoint2.split(':')[1])

    @staticmethod
    def infer_strand(strand:str):
        """ Infer strand """
        if strand == '+':
            return 1
        if strand == '-':
            return -1
        if strand == '.':
            return 0
        raise ValueError(f"Can not infer strand from {strand}")

    def transcript_on_antisense_strand(self, anno:gtf.GenomicAnnotation) -> bool:
        """ Checks whether donor or accepter transcripts are on the antisense
        strand (-) or if the strand can not be interpreted (.) by Arriba. """
        gene_model1 = anno.genes[self.gene_id1]
        gene_model2 = anno.genes[self.gene_id2]
        return self.transcript_strand1 != gene_model1.strand or\
            self.transcript_strand2 != gene_model2.strand

    def get_donor_transcripts(self, anno:gtf.GenomicAnnotation
            ) -> List[gtf.TranscriptAnnotationModel]:
        """ Get all possible donor transcripts """
        pos = self.breakpoint1_position - 1
        return anno.get_transcripts_with_position(self.gene_id1, pos)

    def get_accepter_transcripts(self, anno:gtf.GenomicAnnotation
            ) -> List[gtf.TranscriptAnnotationModel]:
        """ Get all possible donor transcripts """
        pos = self.breakpoint2_position - 1
        return anno.get_transcripts_with_position(self.gene_id2, pos)

    def is_valid(self, min_split_reads1:int, min_split_reads2, confidence:str
            ) -> bool:
        """ Checks if the ArribaRecord is valid """
        return self.split_reads1 >= min_split_reads1 and \
            self.split_reads2 >= min_split_reads2 and \
            self.confidence >= ArribaConfidence(confidence)

    def convert_to_variant_records(self, anno:gtf.GenomicAnnotation,
            genome:dna.DNASeqDict) -> List[seqvar.VariantRecord]:
        """ Convert the Arriba fusion record to a VariantRecord.
        Arriba uses a 1-based coordinate system.

        For Arriba, `breakpoint1` is used to indicate the genomic location of
        the breakpoint of the donor gene, while `breakpoint2` is to that of the
        accepter gene. `direction1` and `direction2` are either 'upstream' or
        'downstream' depends on donor/accepter and its strand. For example,
        for a donor gene on + strand, the value of `direction1` would be
        downstream meaning that the breakpoint is to the downstream side of
        `breakpoint1`. While an acceptor gene on - strand will have the
        `direction2` value being `downstream`. Noted that on genetic
        coordinates, the breakpoint of donor gene will always be 'downstream'
        and that of an accepter gene will always be 'upstream'.

        Args:
            anno (gtf.GenomicAnnotation): genomic annotation loaded from a GTF
                file.
            genome (dna.DNASeqDict): genome sequences loaded from a genome
                FASTA file.

        Return:
            A list of seqvar.VaraintRecord.
        """
        try:
            donor_gene_model = anno.genes[self.gene_id1]
        except KeyError as error:
            raise err.GeneNotFoundError(self.gene_id1) from error
        try:
            accepter_gene_model = anno.genes[self.gene_id2]
        except KeyError as error:
            raise err.GeneNotFoundError(self.gene_id2) from error

        donor_gene_symbol = donor_gene_model.gene_name
        donor_chrom = donor_gene_model.chrom
        left_breakpoint = self.breakpoint1_position
        donor_genome_position = f'{donor_chrom}:{left_breakpoint}:{left_breakpoint}'
        donor_position = anno.coordinate_genomic_to_gene(left_breakpoint - 1, self.gene_id1) + 1
        donor_transcripts = self.get_donor_transcripts(anno)

        accepter_gene_symbol = accepter_gene_model.gene_name
        accepter_chrom = accepter_gene_model.chrom
        right_breakpoint = self.breakpoint2_position
        accepter_genome_position = f'{accepter_chrom}:{right_breakpoint}:{right_breakpoint}'
        accepter_position = anno.coordinate_genomic_to_gene(right_breakpoint - 1, self.gene_id2)
        accepter_transcripts = self.get_accepter_transcripts(anno)

        records = []

        if donor_gene_model.strand == 1:
            ref_seq = genome[donor_chrom].seq[left_breakpoint]
        else:
            ref_seq = genome[donor_chrom]\
                .seq[left_breakpoint:left_breakpoint + 1]\
                .reverse_complement()
            ref_seq = str(ref_seq)

        perms = itertools.product(donor_transcripts, accepter_transcripts)
        for donor_tx, accepter_tx in perms:
            donor_tx_id = donor_tx.transcript.transcript_id
            accepter_tx_id = accepter_tx.transcript.transcript_id

            fusion_id = f'FUSION-{donor_tx_id}:{donor_position}'\
                f'-{accepter_tx_id}:{accepter_position}'

            location = FeatureLocation(
                seqname=self.gene_id1,
                start=donor_position,
                end=donor_position + 1
            )
            attrs = {
                'TRANSCRIPT_ID': donor_tx_id,
                'GENE_SYMBOL': donor_gene_symbol,
                'GENOMIC_POSITION': donor_genome_position,
                'ACCEPTER_GENE_ID': self.gene_id2,
                'ACCEPTER_TRANSCRIPT_ID': accepter_tx_id,
                'ACCEPTER_SYMBOL': accepter_gene_symbol,
                'ACCEPTER_POSITION': accepter_position,
                'ACCEPTER_GENOMIC_POSITION': accepter_genome_position
            }
            record = seqvar.VariantRecord(
                location=location,
                ref=ref_seq,
                alt='<FUSION>',
                _type='Fusion',
                _id=fusion_id,
                attrs=attrs
            )
            records.append(record)
        return records



""" GVF Metadata Info """
from typing import Dict, Union


GVF_METADATA_INFO:Dict[str,Dict[str,Union[int,str]]] = {
    'Base': {
        'TRANSCRIPT_ID': {
            'Number': 1,
            'Type': 'String',
            'Description': 'Transcript ID'
        },
        'GENE_SYMBOL': {
            'Number': 1,
            'Type': 'String',
            'Description': 'Gene Symbol'
        },
        'GENOMIC_POSITION': {
            'Number': 1,
            'Type': 'String',
            'Description': 'Genomic Position'
        }
    },
    'Fusion': {
        'ACCEPTER_GENE_ID': {
            'Number': 1,
            'Type': 'String',
            'Description': "3' Accepter Transcript's Gene ID"
        },
        'ACCEPTER_TRANSCRIPT_ID': {
            'Number': 1,
            'Type': 'String',
            'Description': "3' Accepter Transcript's Transcript ID"
        },
        'ACCEPTER_POSITION': {
            'Number': 1,
            'Type': 'Integer',
            'Description': "Position of the break point of the 3' accepter transcript"
        }
    },
    'Insertion': {
        'DONOR_START': {
            'Number': 1,
            'Type': 'Integer',
            'Description': 'Donor Start Position'
        },
        'DONOR_END': {
            'Number': 1,
            'Type': 'Integer',
            'Description': 'Donor End Position'
        }
    },
    'Deletion': {
        'START': {
            'Number': 1,
            'Type': 'Integer',
            'Description': 'Start Position'
        },
        'END': {
            'Number': 1,
            'Type': 'Integer',
            'Description': 'End Position'
        }
    },
    'Substitution': {
        'START': {
            'Number': 1,
            'Type': 'Integer',
            'Description': 'Start Position'
        },
        'END': {
            'Number': 1,
            'Type': 'Integer',
            'Description': 'End Position'
        },
        'DONOR_START': {
            'Number': 1,
            'Type': 'Integer',
            'Description': 'Donor Start Position'
        },
        'DONOR_END': {
            'Number': 1,
            'Type': 'Integer',
            'Description': 'Donor End Position'
        },
        'COORDINATE': {
            'Number': 1,
            'Type': 'String',
            'Description': 'Coordinate for Insertion or Substitution'
        }
    },
    'circRNA': {
        'OFFSET': {
            'Number': '+',
            'Type': 'Integer',
            'Description': "Offsets of fragments (exons or introns)"
        },
        'LENGTH': {
            'Number': '+',
            'Type': 'Integer',
            'Description': "Lengths of fragments (exons or introns)"
        },
        'INTRON': {
            'Number': '+',
            'Type': 'Integer',
            'Description': "Indices of fragments that are introns"
        }
    }
}

GVF_METADATA_ADDITIONAL:Dict[str,Dict[str,Dict[str,Union[int, str]]]] = {
    'circRNA': {
        'POS': {
            'Description': '"Gene coordinate of circRNA start"'
        }
    }
}



""" Module for sequence variant """
from moPepGen.seqvar.VariantRecord import VariantRecord
from moPepGen.seqvar.VariantRecordWithCoordinate import \
    VariantRecordWithCoordinate
from moPepGen.seqvar import io
from moPepGen.seqvar.GVFMetadata import GVFMetadata
from moPepGen.seqvar.VariantRecordPool import VariantRecordPool
from moPepGen.seqvar.VariantRecordPoolOnDisk import VariantRecordPoolOnDisk, \
    TranscriptionalVariantSeries, VariantRecordPoolOnDiskOpener
from moPepGen.seqvar.VariantRecord import create_variant_sect, create_variant_w2f, \
    create_mnv_from_adjacent, find_mnvs_from_adjacent_variants
from moPepGen.seqvar.SplicingJunction import SpliceJunction, \
    SpliceJunctionTranscriptAlignment
